#!/usr/bin/env python
# coding: utf-8


# ESA (C) 2000-2022
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# colimplot_graph.py


import matplotlib.pyplot as plt
import numpy as np
from astropy.visualization import make_lupton_rgb
import sys
import os
import warnings
from astropy.io import fits
import pysas.pysasplot_utils.pysasplot_utils as sasplt
from matplotlib.colors import LogNorm
import subprocess


def fits_info(fits_file):
    """
    Gathers needed information for the plot from FITS files.

    Args:
        fits_info: the FITS file to evaluate.

    Output:
        (instrument, filter, object, observer, datee, exposure)
        If a result is not found, it will be marked as unknown.
    """

    try:
        with fits.open(fits_file) as ff:
            try:
                instr = ff[0].header['INSTRUME']
            except KeyError:
                instr = 'unknown'
            try:
                filter_info = ff[0].header['FILTER']
            except KeyError:
                filter_info = 'unknown'
            try:
                obj = ff[0].header['OBJECT']
            except KeyError:
                obj = 'unknown'
            try:
                observer = ff[0].header['OBSERVER']
            except KeyError:
                observer = 'unknown'
            try:
                date_obs = ff[0].header['DATE-OBS']
            except KeyError:
                date_obs = 'unknown'
            try:
                exposure = ff[0].header['EXPOSURE']
            except KeyError:
                exposure = 'unknown'
    except FileNotFoundError:
        warnings.warn('Could not open a FITS file to gather info.')
        return ('unknown', 'unknown', 'unknown', 'unknown', 'unknown', 'unknown')

    return (instr, filter_info, obj, observer, date_obs, exposure)


def load_mask(mask):
    """
    Loads the mask from the temporary mask file and makes it readable by NumPy 
    array standards.

    Args:
        mask: the path to the mask temporary file.

    output:
        1 if an error was found. mask_array: an array containing the mask values.
    """

    bool_mask_f = 'bool_mask'

    try:
        with open(mask) as mk:
            lines = mk.readlines()[0]
            lines = lines.replace('\n', '')
            lines = lines.replace('F', '0')
            lines = lines.replace('T', '1')
            with open(bool_mask_f, 'w') as bmk:
                bmk.write(lines)
    except FileNotFoundError:
        warnings.warn('Could not open the mask file.')
        return 1

    os.remove(mask)

    mask_array = np.loadtxt(bool_mask_f)
    mask_array = mask_array.reshape((img_shape[0], img_shape[1]))
#    mask_array = mask_array[::-1]

    os.remove(bool_mask_f)
    return mask_array


def prepare_text(basic_info, img_shape):
    """
    Prepares a text for the figure using data from the first FITS file in list.

    Args:
        basic_info: tuple containing information from the FITS file.
        img_shape: the shape of the 2D arrays forming the RGB bands.

    Ouput:
        text: the formatted text.
    """

    instr, filter_info, obj, observer, date_obs, exposure = basic_info
    text = 'Instrument: {0}\n'.format(instr) + \
            'Filter: {0}\n'.format(filter_info) + \
            'Object: {0}\n'.format(obj) + \
            'Observer: {0}\n'.format(observer) + \
            'Date: {0}\n'.format(date_obs) + \
            'Exposure: {:.1f} seconds\n'.format(exposure) + \
            'Pixel info: X: ({0}, {1}); Y: ({2}, {3})'.format(0, img_shape[0] - 1, 0, img_shape[1] - 1)

    return text


def load_and_reshape(band, img_shape): 
    """
    Loads and reshapes into a 2D array the given file provided by path.

    Args:
        band: the path to the file.
        img_shape: tuple containing the new shape. Compatible with np.reshape.

    Ouput:
        band: the loaded band in a 2D array.
    """

    try:
        band = np.loadtxt(band) 
    except FileNotFoundError:
        raise FileNotFoundError('Could not open {}i. Quitting...'.format(band))    
    band = band.reshape(img_shape) 
#    band = band[::-1] 
    
    return band 


def check_file_path(file_path):
    """
    Gets the information needed to check and evaluate the filename given.

    Args:
        file_path: the path to the file.

    Output:
        the full path, the basename without extension and the extension (if any).
    """

    bsname = os.path.basename(file_path)
    full_path = os.path.abspath(file_path)
    name, extension = os.path.splitext(bsname)
        
    return(full_path, name, extension)


def main(img_shape, filename, device):
    """
    Plotting utility for colimplot.

    Args:
        img_shape: tuple containing the dimensions of the image.
        filename: the output name.
        device: plotting device
    """

    mask_flag = False

    # Opening image
    red = load_and_reshape('red.txt', img_shape)
    green = load_and_reshape('green.txt', img_shape)
    blue = load_and_reshape('blue.txt', img_shape)


    # Searching for mask...
    try:
        mask_array = load_mask('mask')
        mask_flag = True
    except FileNotFoundError:
        print('No mask product found.')
    
    if isinstance(mask_array, int):
        mask_flag = False

    if mask_flag:
        red = np.ma.masked_where(mask_array, red)
        green = np.ma.masked_where(mask_array, green)
        blue = np.ma.masked_where(mask_array, blue)


    rgb = np.dstack((red, green, blue))
    image = make_lupton_rgb(red, green, blue, stretch = 0.1, Q = 0.4)
    f, (ax1, ax2) = plt.subplots(1, 2)

    norm = plt.Normalize(0, np.max(rgb))
    title = subprocess.run(['colimplot', '-v'], capture_output = True, text = True).stdout  
     
    plt.style.use('dark_background')
    image = ax1.imshow(image, origin = 'lower', cmap = 'gray')
# MPL todo junto    image = ax1.imshow(rgb, origin = 'lower')
    #ax1.imshow(mask_array, alpha = 0.1)

    # Axes style
    ax1.set_xlabel('X')
    ax1.set_ylabel('Y')
    ax1.xaxis.label.set_color('white')
    ax1.yaxis.label.set_color('white')
    ax1.set_title('False-Xray colour image.', color = 'white')
    ax1.spines['bottom'].set_color('white')
    ax1.spines['top'].set_color('white')
    ax1.spines['left'].set_color('white')
    ax1.spines['right'].set_color('white')
    ax1.tick_params(axis = 'x', colors = 'white')
    ax1.tick_params(axis = 'y', colors = 'white')

    
    plt.title(title, color = 'white', loc='left', fontsize=6)
    

    # Add text
    try:
        with open('input_files.txt') as in_files:
            default_file = in_files.readlines()[0].strip().rstrip()
            default_file = default_file.replace('\n', '')
        os.remove('input_files.txt')
    except FileNotFoundError:
        warnings.warn('Could not open input_files.txt.')

    basic_info = fits_info(default_file)
    text = prepare_text(basic_info, img_shape)

    ax2.axis('off')
    ax2.text(0,0.5, text)
    ax1.set_title('False-colour X-ray image')     
    
    # Cleaning up...
    os.remove('red.txt')
    os.remove('blue.txt')
    os.remove('green.txt')

    # Either show or save
    if 'XW' in device.upper():
       plt.ion()
       plt.show(block=True)
    else:
       plt.savefig('py' + filename)

    plt.close()

    return None


if __name__ == '__main__':
    img_shape = (int(sys.argv[2]), int(sys.argv[1]))
    filename = sys.argv[3]
    extension = sys.argv[4].replace('/', '.')
    full_path, filename, original_extension = check_file_path(filename)
   
    # RDS - set device explicitly to allow display to terminal
    device = extension
    
    if extension.upper() == original_extension.upper():
        filename = filename + extension
    else:
        # Check extension type for device if not displayed to screen
        if 'XW' not in device.upper():
           warnings.warn('The extension in the name of the file does not match the pgdev parameter. Program will try to using {}.'.format(extension))
           format_flag = sasplt.check_format_compatibility(extension.replace('.', ''))
           if not format_flag:
               warnings.warn('The format {} is not available. Using .png instead.'.format(extension))
               filename = filename + '.png'
           else:
               filename = filename + extension

    main(img_shape, filename, device)
