! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: colimplot_mod
!
! Authors:  Ian Stewart - SSC-LUX
!        :  Masaaki Sakano - SSC-LUX
!
! Description: The task produces image plots, colour-coded to represent the relative intensities in three spectral bands.
!
! Comments:
!
!	TODO:
!	- Access optical filter bandpasses in integrateSpectrum().
!	- Check other promises in 'future dev' in doco.
!	- check to see if band limits are (or should be) stored in the cal. Other issues re energy ranges (overlap with lccorr).
!	- Rationalise the transforms in getColourSaturatedImage(). They should all be such that thermalClr is aligned with the new x axis and purplicity with the new y.
!	- Remove bug errors and checks at some stage.
!	- Clean up comments etc
!	- param 'cutoff': probably better expressed as a fraction of pixels.
!	- implement 'preserveaspect' parameter.
!	- user-entered ref colour triplet? See doco.
!	- enable square-root flux transform and find out why the log one looks so bad.
!	- Get rid of variables and procedure names which have underscores.
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module colimplot_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal
  use param  
  use errorHandling
  use dataFilter
  use dss
!  use caloaldefs
  use metatask_mod ! to get task name and version number.

  use array_utils	! for getBinNumber, maxLoc1d, maskArrayLimits, valueAtHistoFraction
  use constants
  use dal_aux		! for readArrayData
  use dss_aux		! for hasDss
  use dss_ranges_aux	! for readRanges, correctRanges
  use exposure		! for convertPiToEeV
  use rebinners		! for rebinImage.
  use sort_mod
  use ssc_misc		! for fracVal, intToStr
  use warp_utils	! for warp

  use colimplot_common

  implicit none 

  private 

  ! Command-line parameters:
  character(len=20) ::&
    refClrStyle = '',&
    autoWhiteStyle = '',&
    userWhiteStyle = '',&
    negRemovalStyle = '',&
    fluxScalingStyle = '',&
    gainStyle = '',&
    rgbSystemStyle = '',&
    rebinStyle = '',&
    maskStyle = ''
  !character(len=255), pointer ::&
  character(len=255), allocatable ::&
    imageSetName(:)
  character(len=255) ::&
testType = '',&
pgDev = '',&
outFileName = '',&
    maskSetName = ''
  real(single) ::&
    kPixelFrac = 0.0,&
    kFluxFrac = 0.0,&
    whitekeV = 0.0,&
    whiteAngstroms = 0.0,&
    whiteMegaKelvin = 0.0,&
    clrContrast = 0.0,&
    colourKnob(red:blu) = 0.0,&
    userRefColour(red:blu) = 0.0,&
    gain = 0.0,&
    divideNumXbinsBy = 0.0,&
    divideNumYbinsBy = 0.0,&
    bkgLevelFrac = 0.0,&
!    weirdness = 0.0,&		! now def in colimplot_common.f90
!    heat = 0.0,&		! now def in colimplot_common.f90
!    heatSpread = 0.0,&		! now def in colimplot_common.f90
    meanCalcRelFluxCutoff = 0.0,&
    nonMaskGreyLevel = 0.0
  integer(int16) ::&
!    numSuppliedBands,&
    newNumXbins = 0,&
    newNumYbins = 0,&
    numHotPixels = 0
  logical ::&
    rebinImagePlease = .false.,&
    withMask = .false.,&
    preserveAspect = .false.,&
    expandToMask = .false.,&
    doThumbnail = .false.,&
    withFrame = .false.

  ! Other globals:
  type(TaskInfo) :: info ! type from metatask_mod

logical, parameter :: gTestFlag = .false.

!  real(single) :: refThermIndex(-100:100), refThermRGB(-100:100, red:blu)
  integer(int16) ::&
    maxNumBranches, sxi, syi
  real(single), parameter ::&
    plancksConst = 6.6260755e-34,&
    wienConst = 2.897756e-3,&
    mySpeedOfLight = 2.99792458e8
  real(single), parameter ::&
    minVpX = 0.1,&
    maxVpX = 0.9,&
    minVpY = 0.08,&
    maxVpY = 0.92
  real(single) ::&
    gainCorrector
!,&
!    clrGainDistortion
  real, dimension(6) ::&
    trMatrix
  type :: clrValsType
    real(single), dimension(3, 2) :: dimens
    real, dimension(3) :: averageClr
    integer(int32) :: numContainedClrPts
  end type
  character(len=20) ::&
    instrumeStr,&
    filterStr,&
    objectStr,&
    observerStr,&
    dateobsStr,&
    exposureStr,&
    expIdStr
  logical ::&
    notUseableInstrument
  integer, parameter :: defaultNumIntegralSamples = 100

  interface imageRebinner
    module procedure imageRebinnerSingle, imageRebinnerBool
  end interface

  public colimplot

contains 

  !---------------------------------------------------------------------
  subroutine colimplot
  !---------------------------------------------------------------------

    implicit none 

    integer(kind=int16) ::&
      xsize,&
      ysize,&
      oldXsize,&
      oldYsize,&
      clr,&
      i
    integer ::&
      lastClrIndex
    integer ::&
      status,&
      pgopen
    real(kind=single), dimension(:,:,:), allocatable ::&
      tempImage
    real, dimension(:,:), allocatable ::&
      indexImage
    real(kind=single), dimension(:,:,:), pointer ::&
      images
    real(kind=single), dimension(:,:), allocatable ::&
      aveClrValues, tempAveClrValues, netFluxImage
!    real(kind=single), dimension(:,:,:), allocatable ::&
!      rgbImage
    real(kind=single), dimension(:,:,:), pointer ::&
      rgbImage
!    logical(bool), dimension(:,:), pointer ::&
    logical, dimension(:,:), pointer ::&
      mask
        
    logical :: file_exists
    character(4) :: xsize_str, ysize_str
    character(30) :: img_shape
    character(255) :: cmd_text

character(len=255) :: pgFileDev

!integer :: i, xi, yi, clr
!real :: testRgb(red:blu)

!-----------------------------------------------------------------------
! Start:

!testRgb = (/0.2, 0.3, 0.4/)
!print *, maxloc(testRgb)
!return

!gTestFlag = .false.
sxi = 64
syi = 74

    info = getTaskInfo()
    call readParameters

!call plotThermalHues(4)
!return

    call otherInit
!xsize = 300
!ysize = 300
!allocate(images(numSuppliedBands, xsize, ysize))
!call makeRandomImages(images)
!do i = 1, numSuppliedBands
!  imageBandCentres(i) = real(i)
!end do
!averageBandCentre = sum(imageBandCentres) / real(numSuppliedBands)
!call getColourSaturatedImage(images, rgbImage)
!call plotHues(rgbImage)
!return

    call inputChecks(xsize, ysize)
    call readAndRebinImages(images, xsize, ysize, oldXsize, oldYsize)
    call checkNoNegatives(images, xsize, ysize)
    call energyInit

    allocate(netFluxImage(xsize, ysize))
    call getNetFluxImage(images, xsize, ysize, netFluxImage)
    ! After this call netFluxImage contains values which can range from 0 to 1.

    if (withMask) then
      if (maskStyle == 'user') then
        call loadAndRebinMask(mask, xsize, ysize, oldXsize, oldYsize)
      else ! self-mask
        allocate(mask(xsize,ysize))
        mask = netFluxImage > 0.0
      end if
    else
      allocate(mask(xsize,ysize))
      mask = .true.
    end if

!.......................................................................
! Colour and brightness transforms:

    ! Further processing requires that the brightness or net flux information
    ! be handled separately to the colour information. The upcoming calls
    ! getNetFluxImage() and getColourSaturatedImage() perform this separation.
    ! The result is two arrays: the array netFluxImage() containing the
    ! brightness information (just the sum of the normalised image() values)
    ! and the array rgbImage() containing (at present) just the colour
    ! information (rgb values in the range 0 to 1 - note that at least 1 of
    ! the three values per pixel should equal 1 at this stage). The sub
    ! getColourSaturatedImage() actually performs quite a bit more processing
    ! on the colour part of image: mostly this consists of changing the
    ! colours to be either closer or farther from the 'thermal' colour
    ! sequence red-yellow-white-blue, as specified on the command line.

    if (gainStyle == 'auto') then
!      call calcGain(netFluxImage, mask, xsize, ysize)
call calcGain(netFluxImage, mask)
    end if ! A value for gain has already been calculated in readAndRebinImages() from the values of the parameters --gain and --nhotpixels, under the assumption that gainStyle == 'user'. This is now set aside and the gain calculated from the values of the variables kPixelFrac and kFluxFrac.

    ! At this stage the image colour information is still contained in the
    ! array images(), which contains either 2 or 3 input images. These have
    ! been rebinned if that was requested on the command line but not
    ! otherwise scaled as yet. Each image pixel is therefore associated with
    ! either 2 or 3 numbers, one from each input image. The first thing to do
    ! is to normalise these values by dividing each pair or triplet of values
    ! per pixel by a reference pair or triplet. This eliminates biases due to
    ! the sizes and positions of the filter bandpasses in the OM case and
    ! selected energy bands in the EPIC case.

!    call divideByRefColour(images, netFluxImage, mask, xsize, ysize)
call divideByRefColour(images, netFluxImage, mask)

!    allocate(rgbImage(xsize, ysize, red:blu))
!    call getColourSaturatedImage_old(images, xsize, ysize, rgbImage)
call getColourSaturatedImage(images, rgbImage) ! also deallocates images().
deallocate(bbRefColour, imageBandCentres)	! Global allocatable vars.
deallocate(multiplet)	! Global pointer used in getThermalClrAndPurplicity and
  ! getThermalIndices but neither the functions is used.

!    deallocate(images)
!call plotHues(rgbImage)
!return

    ! The next call transforms the net flux scale to logarithmic, square root
    ! etc as desired.

!    call netFluxTransform(netFluxImage, mask, xsize, ysize)
call netFluxTransform(netFluxImage, mask)

    ! Most of the transformations are now done. Although it is convenient to
    ! retain the array netFluxImage() for a few more steps, after the
    ! following call to recombineImageInfo() it is essentially redundant,
    ! as the brightness information is recombined with the colour info.
    ! The array rgbImage() will then contain the recombined information.
    ! Note that the rgbImage() values will still be within the range 0 to 1,
    ! but now any combination will be allowed (previously 1 at least of the
    ! three values for each pixel had to = 1.0, since hues were being
    ! manipulated at that stage).

    call recombineImageInfo(rgbImage, netFluxImage, xsize, ysize)

    ! The rgbImage() values will next be scaled so as to saturate brighter
    ! parts of the image, if this is requested on the command line.

    call netFluxSaturate(rgbImage, netFluxImage, xsize, ysize)

    deallocate(netFluxImage)

    ! Finally, the red, green or blue values can be separately amplified
    ! or attenuated if desired:

!    call finalColourTweak(rgbImage, mask, xsize, ysize)
do clr = red, blu
  where (.not. mask) rgbImage(:, :, clr) = nonMaskGreyLevel
end do


!.......................................................................
!.......................................................................
! Arranging the output:

    ! Note that the following three output formats are _not_ done via pgplot
    ! output. This is deliberate policy, designed to thereby side-step the
    ! 8-bit colour descriptor of pgplot and its consequent limit of 256
    ! colours. By writing the .cps file directly we can use as many colours
    ! as we want.

    if (trim(pgDev) == '/cps') then
!      call drawPostscript(rgbImage, mask, xsize, ysize, oldXsize&
!call drawPostscript(rgbImage, logical(mask, bool), oldXsize&
call drawPostscript(rgbImage, mask, oldXsize&
      , oldYsize, outFileName)
      deallocate(rgbImage)
      deallocate(imageSetName)
  return

    else if (trim(pgDev) == '/fits') then
!      call saveToFits(rgbImage, mask, xsize, ysize, outFileName)
!call saveToFits(rgbImage, logical(mask, bool), outFileName)
call saveToFits(rgbImage, mask, outFileName)
      deallocate(rgbImage)
      deallocate(imageSetName)
  return

    else if (trim(pgDev) == '/ppm') then
!      call saveToPPM(rgbImage, mask, xsize, ysize, outFileName)
!call saveToPPM(rgbImage, logical(mask, bool), outFileName)
call saveToPPM(rgbImage, mask, outFileName)
      deallocate(rgbImage)
      deallocate(imageSetName)
  return
    end if

    ! The remainder of the program concerns output to pgplot where that is
    ! inevitable. The following routines are designed to deal as best as
    ! possible with the restrictions that pgplot puts on the maximum number
    ! of colours that can be displayed at the same time. Because of this
    ! restriction it is necessary to assign or bin the (possibly large)
    ! number of colour combinations in the image to the much more restricted
    ! set of pgplot colour indices. This should be done as intelligently as
    ! possible so that the resulting image looks acceptable.

    if (trim(pgDev) == '/ps' .or. trim(pgDev) == '/gif') then
      pgFileDev = trim(outFileName)//trim(pgDev)
    else
      pgFileDev = trim(pgDev)
    end if

    status = pgopen(pgFileDev)
if (status .le. 0) call error('badPgplotDevice', 'PGPLOT did not recognise &
&the device '//trim(pgDev)//' specified by parameter pgdev.')
    call pgqcol(minClrIndex, maxClrIndex)
    minClrIndex = 2
    maxNumClrIndices = 1 + maxClrIndex - minClrIndex

    allocate(tempAveClrValues(minClrIndex:maxClrIndex, red:blu))
    allocate(aveClrValues(minClrIndex:maxClrIndex, red:blu))
    allocate(indexImage(xsize, ysize))
    allocate(tempImage(xsize, ysize, red:blu))

    if (rgbSystemStyle == 'polar') then
      call toPolarCoord(rgbImage, tempImage, xsize, ysize)
    else
      tempImage = rgbImage
    end if

    maxNumBranches = 2 * maxNumClrIndices - 1
    call assignColourValues(tempImage, xsize, ysize, maxNumBranches&
    , indexImage, lastClrIndex, tempAveClrValues)
    deallocate(tempImage)

    if (rgbSystemStyle == 'polar') then
      call fromPolarCoord(tempAveClrValues, aveClrValues)
    else
      aveClrValues = tempAveClrValues
    end if

!.......................................................................
! Now plot:

    call outputToPgplot(aveClrValues, minClrIndex, maxClrIndex&
!    , lastClrIndex, indexImage, mask, xsize, ysize)
, lastClrIndex, mask, indexImage)


    open(33, file = 'mask', status = 'new')
    write(33, *) mask
    close(33)

    open(33, file = 'red.txt', status = 'new')
    write(33, *) rgbImage(:, :, 1)
    close(33)

    open(33, file = 'green.txt', status = 'new')
    write(33, *) rgbImage(:, :, 2)
    close(33)

    open(33, file = 'blue.txt', status = 'new')
    write(33, *) rgbImage(:, :, 3)
    close(33)


!    write(*, *) ysize, xsize, size(mask, 1), size(mask, 2), size(rgbImage, 1), size(rgbImage, 2), size(rgbImage, 3)

    deallocate(rgbImage, mask)
    deallocate(imageSetName)
   
    open(33, file = 'input_files.txt', status = 'new')
    do i = 1, numSuppliedBands     
        write(33, '(a)') trim(stringParameter('insets', i - 1))
    end do
    close(33)

    if (xsize .lt. 1000 .and. xsize .gt. 99) then
            write(xsize_str, '(a, I3)') '0', xsize
    elseif (xsize .lt. 100 .and. xsize .gt. 9) then
            write(xsize_str, '(a, I2)') '00', xsize
    elseif (xsize .lt. 10) then
            write(xsize_str, '(a, I1)') '000', xsize
    else
            write(xsize_str, '(I4)') xsize
    endif

    if (ysize .lt. 1000 .and. ysize .gt. 99) then
            write(ysize_str, '(a, I3)') '0', ysize
    elseif (ysize .lt. 100 .and. ysize .gt. 9) then
            write(ysize_str, '(a, I2)') '00', ysize
    elseif (ysize .lt. 10) then
            write(ysize_str, '(a, I1)') '000', ysize
    else
            write(ysize_str, '(I4)') ysize
    endif

    write(cmd_text, '(a, a, a, a, a, a, a, a)') 'colimplot_graph.py ', xsize_str, ' ', ysize_str, ' ', &
            trim(outFileName), ' ', trim(pgDev)

  write(*,*)cmd_text ! RDS debug
  call execute_command_line(cmd_text)

!  Deletes temporary files in case the Python script runs into an error.
  inquire(file = 'red.txt', exist = file_exists)
  if (file_exists) then
          open(22, file = 'red.txt', status = 'old')
          close(22, status = 'delete')
  end if
  inquire(file = 'green.txt', exist = file_exists)
  if (file_exists) then
          open(22, file = 'green.txt', status = 'old')
          close(22, status = 'delete')
  end if
  inquire(file = 'blue.txt', exist = file_exists)
  if (file_exists) then
          open(22, file = 'blue.txt', status = 'old')
          close(22, status = 'delete')
  end if
  inquire(file = 'mask', exist = file_exists)
  if (file_exists) then
          open(22, file = 'mask', status = 'old')
          close(22, status = 'delete')
  end if

  inquire(file = 'input_files.txt', exist = file_exists)
  if (file_exists) then
          open(22, file = 'input_files.txt', status = 'old')
          close(22, status = 'delete')
  end if

  end subroutine colimplot

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: i, numUserRefColours

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading parameters.')

    numSuppliedBands = parameterCount('insets')
    if (numSuppliedBands < 1)&
       call error('noInsets', 'Command-line parameter insets is mandatory.')
    allocate(imageSetName(numSuppliedBands))
    do i = 1, numSuppliedBands
      ! imageSetName(i) = trim(stringParameter('insets', i - 1))
      imageSetName(i) = stringParameter('insets', i - 1)
      imageSetName(i) = trim(imageSetName(i))
    end do

    withMask = booleanParameter('withmask')
    nonMaskGreyLevel = 1.0
    if (withMask) then
      maskStyle = trim(stringParameter('maskstyle'))
      if (maskStyle == 'user') maskSetName = trim(stringParameter('maskset'))
      nonMaskGreyLevel = realParameter('greylevel')
      expandToMask = booleanParameter('expandtomask')
    end if

    pgDev = trim(stringParameter('pgdev'))
    outFileName = trim(stringParameter('plotfile'))
    withFrame = booleanParameter('withframe')

    rebinImagePlease = booleanParameter('rebinimage')
    if (rebinImagePlease) then
      rebinStyle = trim(stringParameter('rebinstyle'))
      select case(rebinStyle)
      case('tonumber')
        newNumXbins = intParameter('newnxbins')
        newNumYbins = intParameter('newnybins')

      case('divideby')
        divideNumXbinsBy = realParameter('dividexby')
        divideNumYbinsBy = realParameter('divideyby')

      case default
call error('badRebinStyle', "Unrecognised value of --rebinstyle.")
      end select
    end if

    negRemovalStyle = trim(stringParameter('negremovalstyle')) ! offset or truncate

    gainStyle = trim(stringParameter('gainstyle')) ! user or auto
    select case(gainStyle)
    case('auto')
      kPixelFrac = realParameter('pixelfraction')
      kFluxFrac = realParameter('tofluxfraction')

    case('user')
      gain = realParameter('usergain')

    case default
call error('badGainStyle', "Unrecognised value of --gainstyle.")
    end select

    numHotPixels = intParameter('nhotpixels')

    refClrStyle = trim(stringParameter('refclrstyle'))
    select case(refClrStyle)
    case('internal')
      autoWhiteStyle = trim(stringParameter('internalrefstyle'))
      meanCalcRelFluxCutoff = realParameter('cutoff')

    case('bb')
      userWhiteStyle = trim(stringParameter('bbrefstyle'))
      select case(userWhiteStyle)
      case('kev')
        whitekeV = realParameter('whitekev')

      case('angstroms')
        whiteAngstroms = realParameter('whiteangstroms')

      case('kelvin')
        whiteMegaKelvin = realParameter('whitemegakelvin')

      case default
call error('badBbRefStyle', "Unrecognised value of --bbrefstyle.")
      end select

    case('user')
      numUserRefColours = parameterCount('refclr')
if (numUserRefColours .ne. 3) call error('',"")
      userRefColour(red) = realParameter('refclr', 0)
      userRefColour(grn) = realParameter('refclr', 1)
      userRefColour(blu) = realParameter('refclr', 2)

    case('none')
    case default
call error('badRefClrStyle', "Unrecognised value of --refclrstyle.")
    end select

    weirdness = realParameter('weirdness')
    heat = realParameter('heat')
    heatSpread = realParameter('heatspread')
    rgbSystemStyle = trim(stringParameter('rgbsysstyle')) ! cartesian, polar


!    fluxScalingStyle = trim(stringParameter('fluxtransformstyle')) ! log, lin, equ
fluxScalingStyle = 'lin'
!    bkgLevelFrac = realParameter('bkglevel')
!if (bkgLevelFrac >= 1.0) call error('bkgFracTooHigh', 'You &
!&specified bkglevel=1.0. Sorry, only values < 1.0 are allowed.')

!    clrContrast = realParameter('clrcontrast')
!    colourKnob(red) = realParameter('redamp')
!    colourKnob(grn) = realParameter('greenamp')
!    colourKnob(blu) = realParameter('blueamp')
colourKnob = 0.0

!doThumbnail = booleanParameter('thumbnail')
doThumbnail = .false.
!    preserveAspect = booleanParameter('preserveaspect')
preserveAspect = .false.

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine otherInit
  !---------------------------------------------------------------------

    implicit none

    integer(int16) :: i
!logical, parameter :: testFlag = .true.

    !-------------------------------------------------------------------
    ! Start:

    write(errStr, '("Number of supplied bands: ",i2)')  numSuppliedBands
    call message(AppMsg, Verbose)
    allocate(imageBandCentres(numSuppliedBands), multiplet(numSuppliedBands)&
    , bbRefColour(numSuppliedBands))
if (gTestFlag) print *, 'size of multiplet (otherInit) = ',size(multiplet)

    trMatrix(1) = 0.0
    trMatrix(2) = 1.0
    trMatrix(3) = 0.0
    trMatrix(4) = 0.0
    trMatrix(5) = 0.0
    trMatrix(6) = 1.0

    refThermRGB(-100, :) = (/1.00000, 0.00000, 0.00000/)
    refThermRGB( -99, :) = (/1.00000, 0.01133, 0.00000/)
    refThermRGB( -98, :) = (/1.00000, 0.02290, 0.00001/)
    refThermRGB( -97, :) = (/1.00000, 0.03469, 0.00004/)
    refThermRGB( -96, :) = (/1.00000, 0.04668, 0.00011/)
    refThermRGB( -95, :) = (/1.00000, 0.05886, 0.00021/)
    refThermRGB( -94, :) = (/1.00000, 0.07119, 0.00037/)
    refThermRGB( -93, :) = (/1.00000, 0.08365, 0.00060/)
    refThermRGB( -92, :) = (/1.00000, 0.09622, 0.00091/)
    refThermRGB( -91, :) = (/1.00000, 0.10887, 0.00132/)
    refThermRGB( -90, :) = (/1.00000, 0.12157, 0.00184/)
    refThermRGB( -89, :) = (/1.00000, 0.13428, 0.00249/)
    refThermRGB( -88, :) = (/1.00000, 0.14700, 0.00327/)
    refThermRGB( -87, :) = (/1.00000, 0.15969, 0.00420/)
    refThermRGB( -86, :) = (/1.00000, 0.17232, 0.00528/)
    refThermRGB( -85, :) = (/1.00000, 0.18489, 0.00653/)
    refThermRGB( -84, :) = (/1.00000, 0.19738, 0.00795/)
    refThermRGB( -83, :) = (/1.00000, 0.20976, 0.00956/)
    refThermRGB( -82, :) = (/1.00000, 0.22204, 0.01135/)
    refThermRGB( -81, :) = (/1.00000, 0.23419, 0.01333/)
    refThermRGB( -80, :) = (/1.00000, 0.24622, 0.01551/)
    refThermRGB( -79, :) = (/1.00000, 0.25811, 0.01788/)
    refThermRGB( -78, :) = (/1.00000, 0.26987, 0.02045/)
    refThermRGB( -77, :) = (/1.00000, 0.28149, 0.02323/)
    refThermRGB( -76, :) = (/1.00000, 0.29297, 0.02620/)
    refThermRGB( -75, :) = (/1.00000, 0.30431, 0.02939/)
    refThermRGB( -74, :) = (/1.00000, 0.31552, 0.03277/)
    refThermRGB( -73, :) = (/1.00000, 0.32659, 0.03636/)
    refThermRGB( -72, :) = (/1.00000, 0.33753, 0.04015/)
    refThermRGB( -71, :) = (/1.00000, 0.34834, 0.04415/)
    refThermRGB( -70, :) = (/1.00000, 0.35903, 0.04835/)
    refThermRGB( -69, :) = (/1.00000, 0.36959, 0.05276/)
    refThermRGB( -68, :) = (/1.00000, 0.38004, 0.05737/)
    refThermRGB( -67, :) = (/1.00000, 0.39038, 0.06218/)
    refThermRGB( -66, :) = (/1.00000, 0.40061, 0.06721/)
    refThermRGB( -65, :) = (/1.00000, 0.41074, 0.07243/)
    refThermRGB( -64, :) = (/1.00000, 0.42077, 0.07787/)
    refThermRGB( -63, :) = (/1.00000, 0.43070, 0.08351/)
    refThermRGB( -62, :) = (/1.00000, 0.44054, 0.08935/)
    refThermRGB( -61, :) = (/1.00000, 0.45030, 0.09541/)
    refThermRGB( -60, :) = (/1.00000, 0.45997, 0.10167/)
    refThermRGB( -59, :) = (/1.00000, 0.46956, 0.10814/)
    refThermRGB( -58, :) = (/1.00000, 0.47909, 0.11482/)
    refThermRGB( -57, :) = (/1.00000, 0.48853, 0.12172/)
    refThermRGB( -56, :) = (/1.00000, 0.49792, 0.12884/)
    refThermRGB( -55, :) = (/1.00000, 0.50724, 0.13616/)
    refThermRGB( -54, :) = (/1.00000, 0.51650, 0.14370/)
    refThermRGB( -53, :) = (/1.00000, 0.52570, 0.15147/)
    refThermRGB( -52, :) = (/1.00000, 0.53485, 0.15946/)
    refThermRGB( -51, :) = (/1.00000, 0.54396, 0.16767/)
    refThermRGB( -50, :) = (/1.00000, 0.55301, 0.17610/)
    refThermRGB( -49, :) = (/1.00000, 0.56202, 0.18477/)
    refThermRGB( -48, :) = (/1.00000, 0.57100, 0.19367/)
    refThermRGB( -47, :) = (/1.00000, 0.57993, 0.20280/)
    refThermRGB( -46, :) = (/1.00000, 0.58883, 0.21217/)
    refThermRGB( -45, :) = (/1.00000, 0.59770, 0.22179/)
    refThermRGB( -44, :) = (/1.00000, 0.60654, 0.23165/)
    refThermRGB( -43, :) = (/1.00000, 0.61535, 0.24176/)
    refThermRGB( -42, :) = (/1.00000, 0.62414, 0.25212/)
    refThermRGB( -41, :) = (/1.00000, 0.63290, 0.26273/)
    refThermRGB( -40, :) = (/1.00000, 0.64165, 0.27361/)
    refThermRGB( -39, :) = (/1.00000, 0.65038, 0.28475/)
    refThermRGB( -38, :) = (/1.00000, 0.65910, 0.29616/)
    refThermRGB( -37, :) = (/1.00000, 0.66780, 0.30784/)
    refThermRGB( -36, :) = (/1.00000, 0.67649, 0.31982/)
    refThermRGB( -35, :) = (/1.00000, 0.68518, 0.33207/)
    refThermRGB( -34, :) = (/1.00000, 0.69386, 0.34460/)
    refThermRGB( -33, :) = (/1.00000, 0.70254, 0.35745/)
    refThermRGB( -32, :) = (/1.00000, 0.71121, 0.37058/)
    refThermRGB( -31, :) = (/1.00000, 0.71989, 0.38403/)
    refThermRGB( -30, :) = (/1.00000, 0.72857, 0.39779/)
    refThermRGB( -29, :) = (/1.00000, 0.73725, 0.41187/)
    refThermRGB( -28, :) = (/1.00000, 0.74594, 0.42628/)
    refThermRGB( -27, :) = (/1.00000, 0.75464, 0.44101/)
    refThermRGB( -26, :) = (/1.00000, 0.76335, 0.45611/)
    refThermRGB( -25, :) = (/1.00000, 0.77207, 0.47152/)
    refThermRGB( -24, :) = (/1.00000, 0.78081, 0.48733/)
    refThermRGB( -23, :) = (/1.00000, 0.78956, 0.50347/)
    refThermRGB( -22, :) = (/1.00000, 0.79833, 0.52001/)
    refThermRGB( -21, :) = (/1.00000, 0.80713, 0.53691/)
    refThermRGB( -20, :) = (/1.00000, 0.81594, 0.55423/)
    refThermRGB( -19, :) = (/1.00000, 0.82478, 0.57192/)
    refThermRGB( -18, :) = (/1.00000, 0.83365, 0.59004/)
    refThermRGB( -17, :) = (/1.00000, 0.84254, 0.60857/)
    refThermRGB( -16, :) = (/1.00000, 0.85146, 0.62754/)
    refThermRGB( -15, :) = (/1.00000, 0.86042, 0.64695/)
    refThermRGB( -14, :) = (/1.00000, 0.86941, 0.66680/)
    refThermRGB( -13, :) = (/1.00000, 0.87844, 0.68714/)
    refThermRGB( -12, :) = (/1.00000, 0.88750, 0.70793/)
    refThermRGB( -11, :) = (/1.00000, 0.89661, 0.72924/)
    refThermRGB( -10, :) = (/1.00000, 0.90575, 0.75104/)
    refThermRGB(  -9, :) = (/1.00000, 0.91494, 0.77335/)
    refThermRGB(  -8, :) = (/1.00000, 0.92418, 0.79621/)
    refThermRGB(  -7, :) = (/1.00000, 0.93346, 0.81958/)
    refThermRGB(  -6, :) = (/1.00000, 0.94280, 0.84356/)
    refThermRGB(  -5, :) = (/1.00000, 0.95219, 0.86809/)
    refThermRGB(  -4, :) = (/1.00000, 0.96163, 0.89321/)
    refThermRGB(  -3, :) = (/1.00000, 0.97113, 0.91897/)
    refThermRGB(  -2, :) = (/1.00000, 0.98069, 0.94531/)
    refThermRGB(  -1, :) = (/1.00000, 0.99032, 0.97235/)
    refThermRGB(   0, :) = (/1.00000, 1.00000, 1.00000/)
    refThermRGB(   1, :) = (/0.96953, 0.97998, 1.00000/)
    refThermRGB(   2, :) = (/0.94004, 0.96042, 1.00000/)
    refThermRGB(   3, :) = (/0.91156, 0.94133, 1.00000/)
    refThermRGB(   4, :) = (/0.88399, 0.92266, 1.00000/)
    refThermRGB(   5, :) = (/0.85737, 0.90445, 1.00000/)
    refThermRGB(   6, :) = (/0.83156, 0.88661, 1.00000/)
    refThermRGB(   7, :) = (/0.80664, 0.86920, 1.00000/)
    refThermRGB(   8, :) = (/0.78245, 0.85214, 1.00000/)
    refThermRGB(   9, :) = (/0.75909, 0.83548, 1.00000/)
    refThermRGB(  10, :) = (/0.73641, 0.81915, 1.00000/)
    refThermRGB(  11, :) = (/0.71445, 0.80317, 1.00000/)
    refThermRGB(  12, :) = (/0.69316, 0.78752, 1.00000/)
    refThermRGB(  13, :) = (/0.67250, 0.77219, 1.00000/)
    refThermRGB(  14, :) = (/0.65249, 0.75717, 1.00000/)
    refThermRGB(  15, :) = (/0.63303, 0.74243, 1.00000/)
    refThermRGB(  16, :) = (/0.61420, 0.72801, 1.00000/)
    refThermRGB(  17, :) = (/0.59588, 0.71384, 1.00000/)
    refThermRGB(  18, :) = (/0.57811, 0.69996, 1.00000/)
    refThermRGB(  19, :) = (/0.56085, 0.68633, 1.00000/)
    refThermRGB(  20, :) = (/0.54405, 0.67294, 1.00000/)
    refThermRGB(  21, :) = (/0.52778, 0.65982, 1.00000/)
    refThermRGB(  22, :) = (/0.51193, 0.64692, 1.00000/)
    refThermRGB(  23, :) = (/0.49652, 0.63425, 1.00000/)
    refThermRGB(  24, :) = (/0.48156, 0.62181, 1.00000/)
    refThermRGB(  25, :) = (/0.46699, 0.60958, 1.00000/)
    refThermRGB(  26, :) = (/0.45282, 0.59755, 1.00000/)
    refThermRGB(  27, :) = (/0.43905, 0.58574, 1.00000/)
    refThermRGB(  28, :) = (/0.42563, 0.57411, 1.00000/)
    refThermRGB(  29, :) = (/0.41257, 0.56267, 1.00000/)
    refThermRGB(  30, :) = (/0.39987, 0.55143, 1.00000/)
    refThermRGB(  31, :) = (/0.38750, 0.54036, 1.00000/)
    refThermRGB(  32, :) = (/0.37545, 0.52946, 1.00000/)
    refThermRGB(  33, :) = (/0.36370, 0.51873, 1.00000/)
    refThermRGB(  34, :) = (/0.35228, 0.50817, 1.00000/)
    refThermRGB(  35, :) = (/0.34115, 0.49777, 1.00000/)
    refThermRGB(  36, :) = (/0.33030, 0.48752, 1.00000/)
    refThermRGB(  37, :) = (/0.31972, 0.47742, 1.00000/)
    refThermRGB(  38, :) = (/0.30941, 0.46746, 1.00000/)
    refThermRGB(  39, :) = (/0.29936, 0.45764, 1.00000/)
    refThermRGB(  40, :) = (/0.28957, 0.44797, 1.00000/)
    refThermRGB(  41, :) = (/0.28002, 0.43843, 1.00000/)
    refThermRGB(  42, :) = (/0.27071, 0.42901, 1.00000/)
    refThermRGB(  43, :) = (/0.26163, 0.41973, 1.00000/)
    refThermRGB(  44, :) = (/0.25277, 0.41056, 1.00000/)
    refThermRGB(  45, :) = (/0.24414, 0.40152, 1.00000/)
    refThermRGB(  46, :) = (/0.23571, 0.39259, 1.00000/)
    refThermRGB(  47, :) = (/0.22749, 0.38377, 1.00000/)
    refThermRGB(  48, :) = (/0.21948, 0.37506, 1.00000/)
    refThermRGB(  49, :) = (/0.21166, 0.36646, 1.00000/)
    refThermRGB(  50, :) = (/0.20403, 0.35796, 1.00000/)
    refThermRGB(  51, :) = (/0.19658, 0.34956, 1.00000/)
    refThermRGB(  52, :) = (/0.18931, 0.34126, 1.00000/)
    refThermRGB(  53, :) = (/0.18222, 0.33305, 1.00000/)
    refThermRGB(  54, :) = (/0.17530, 0.32493, 1.00000/)
    refThermRGB(  55, :) = (/0.16856, 0.31691, 1.00000/)
    refThermRGB(  56, :) = (/0.16197, 0.30897, 1.00000/)
    refThermRGB(  57, :) = (/0.15554, 0.30111, 1.00000/)
    refThermRGB(  58, :) = (/0.14927, 0.29333, 1.00000/)
    refThermRGB(  59, :) = (/0.14315, 0.28563, 1.00000/)
    refThermRGB(  60, :) = (/0.13718, 0.27801, 1.00000/)
    refThermRGB(  61, :) = (/0.13136, 0.27045, 1.00000/)
    refThermRGB(  62, :) = (/0.12567, 0.26297, 1.00000/)
    refThermRGB(  63, :) = (/0.12013, 0.25556, 1.00000/)
    refThermRGB(  64, :) = (/0.11472, 0.24822, 1.00000/)
    refThermRGB(  65, :) = (/0.10945, 0.24093, 1.00000/)
    refThermRGB(  66, :) = (/0.10431, 0.23371, 1.00000/)
    refThermRGB(  67, :) = (/0.09929, 0.22655, 1.00000/)
    refThermRGB(  68, :) = (/0.09441, 0.21945, 1.00000/)
    refThermRGB(  69, :) = (/0.08964, 0.21240, 1.00000/)
    refThermRGB(  70, :) = (/0.08500, 0.20539, 1.00000/)
    refThermRGB(  71, :) = (/0.08048, 0.19844, 1.00000/)
    refThermRGB(  72, :) = (/0.07608, 0.19154, 1.00000/)
    refThermRGB(  73, :) = (/0.07179, 0.18468, 1.00000/)
    refThermRGB(  74, :) = (/0.06761, 0.17786, 1.00000/)
    refThermRGB(  75, :) = (/0.06355, 0.17108, 1.00000/)
    refThermRGB(  76, :) = (/0.05961, 0.16434, 1.00000/)
    refThermRGB(  77, :) = (/0.05577, 0.15763, 1.00000/)
    refThermRGB(  78, :) = (/0.05204, 0.15095, 1.00000/)
    refThermRGB(  79, :) = (/0.04842, 0.14430, 1.00000/)
    refThermRGB(  80, :) = (/0.04491, 0.13767, 1.00000/)
    refThermRGB(  81, :) = (/0.04150, 0.13107, 1.00000/)
    refThermRGB(  82, :) = (/0.03820, 0.12448, 1.00000/)
    refThermRGB(  83, :) = (/0.03501, 0.11790, 1.00000/)
    refThermRGB(  84, :) = (/0.03193, 0.11134, 1.00000/)
    refThermRGB(  85, :) = (/0.02895, 0.10478, 1.00000/)
    refThermRGB(  86, :) = (/0.02608, 0.09822, 1.00000/)
    refThermRGB(  87, :) = (/0.02332, 0.09165, 1.00000/)
    refThermRGB(  88, :) = (/0.02067, 0.08507, 1.00000/)
    refThermRGB(  89, :) = (/0.01814, 0.07847, 1.00000/)
    refThermRGB(  90, :) = (/0.01572, 0.07185, 1.00000/)
    refThermRGB(  91, :) = (/0.01342, 0.06519, 1.00000/)
    refThermRGB(  92, :) = (/0.01125, 0.05849, 1.00000/)
    refThermRGB(  93, :) = (/0.00920, 0.05172, 1.00000/)
    refThermRGB(  94, :) = (/0.00730, 0.04487, 1.00000/)
    refThermRGB(  95, :) = (/0.00554, 0.03793, 1.00000/)
    refThermRGB(  96, :) = (/0.00396, 0.03086, 1.00000/)
    refThermRGB(  97, :) = (/0.00255, 0.02363, 1.00000/)
    refThermRGB(  98, :) = (/0.00137, 0.01617, 1.00000/)
    refThermRGB(  99, :) = (/0.00046, 0.00840, 1.00000/)
    refThermRGB( 100, :) = (/0.00000, 0.00000, 1.00000/)

    do i = -100, 100
      refThermIndex(i) = i / 100.0
    end do

  end subroutine otherInit

  !---------------------------------------------------------------------
  subroutine inputChecks(xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(out) :: xsize, ysize
    type(DataSetT) ::&
      inSet
    type(ArrayT) ::&
      imageArray
    integer, dimension(:), pointer ::&
      dimVector
    !integer(kind=int16), dimension(numSuppliedBands) ::&	! numSuppliedBands is non-parameter global variable.
    integer(kind=int16), allocatable ::&
      tempXsize(:),&
      tempYsize(:)
    integer(kind=int16) :: band
    character(len=20) ::&
      tempInstrumName(numSuppliedBands)		! Should be allocatable????

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Checking input.')

    allocate(tempXsize(numSuppliedBands), tempYsize(numSuppliedBands))
    tempInstrumName = '' ! vector op, default.
    do band = 1, numSuppliedBands
      inSet = dataSet(imageSetName(band), READ)
      if (hasAttribute(inSet, 'INSTRUME')) then
        tempInstrumName(band) = trim(stringAttribute(inSet, 'INSTRUME'))
      end if
      if (trim(tempInstrumName(band)) .ne. tempInstrumName(1)) then
        call warning('badInstrumeKwd', 'The images are from different &
        &instruments.')
      end if

      imageArray = array(inSet, 0, READ)
      dimVector => dimensions(imageArray)
      tempXsize(band) = dimVector(0)
      tempYsize(band) = dimVector(1)

      if (tempXsize(band) .ne. tempXsize(1)&
      .or. tempYsize(band) .ne. tempYsize(1)) then
        call warning('differentSizes', 'The three images are not all of &
        &the same size. The small ones are being padded to the maximum sizes.')
      end if

      call release(inSet)
    end do

    instrumeStr = trim(tempInstrumName(1))
    if (refClrStyle == 'bb') then
      if (instrumeStr == '') then
call error('noInstrumKwd', "The input images have no INSTRUME &
&keyword. Without this keyword, it is impossible to calculate a thermal &
&equivalent reference colour. Use refclrstyle='internal' or 'none'.")
      end if
    end if
!print *, 'instrumeStr, instIdToString(EPN):'
!call stringDump(trim(instrumeStr))
!call stringDump(instIdToString(EPN))
!call error('','') 
    if (instIdToString(EMOS1) .ne. trim(instrumeStr)&
    .and. instIdToString(EMOS2) .ne. trim(instrumeStr)&
    .and. instIdToString(EPN) .ne. trim(instrumeStr)&
    .and. instIdToString(OM) .ne. trim(instrumeStr)) then
      notUseableInstrument = .true.
      if (refClrStyle == 'bb') then
call error('badInstrum', "A thermal equivalent reference colour can only &
&be calculated for EPIC or OM instruments. The instrument is not of that &
&set. To obtain output for this instrument you should set &
&refclrstyle='internal' or 'none'.")
      end if

    else
      notUseableInstrument = .false.
      instrument = instIdToInt(instrumeStr)
      call CAL_setState(instrument = instrument)
      call message(AppMsg, Verbose, 'Set state of CAL for '&
      //trim(instrumeStr))
    end if


    xsize = tempXsize(1)
    ysize = tempYsize(1)
    do band = 2, numSuppliedBands
      xsize = max(xsize, tempXsize(band))
      ysize = max(ysize, tempYsize(band))
    end do

    if (rebinImagePlease) then
      if (rebinStyle == 'divideby') then
        newNumXbins = int(xsize / divideNumXbinsBy, int16)
        newNumYbins = int(ysize / divideNumYbinsBy, int16)
      end if

      if (newNumXbins > xsize) then
!***call error('tooManyXbins', 'The requested number of image x-axis bins is &
!&larger than the actual number', xsize)
      end if
      if (newNumYbins > ysize) then
!***call error('tooManyYbins', 'The requested number of image y-axis bins is &
!&larger than the actual number', ysize)
      end if
    end if

    ! Get other attributes to put on postscript output:
    if (notUseableInstrument) then
      instrumeStr = 'No keyword found.'
    end if
    inSet = dataSet(imageSetName(1), READ)
    if (hasAttribute(inSet, 'EXP_ID')) then
      expIdStr = stringAttribute(inSet, 'EXP_ID')
    else
      expIdStr = 'No keyword found.'
    end if
    call release(inSet)
!    if (hasAttribute(inSet, 'OBJECT')) then
!      objectStr = stringAttribute(inSet, 'OBJECT')
!    else
!      objectStr = 'No keyword found.'
!    end if
!    if (hasAttribute(inSet, 'OBSERVER')) then
!      observerStr = stringAttribute(inSet, 'OBSERVER')
!    else
!      observerStr = 'No keyword found.'
!    end if

  end subroutine inputChecks

  !---------------------------------------------------------------------
  subroutine readAndRebinImages(images, xsize, ysize, oldXsize, oldYsize)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(inout) :: xsize, ysize
    integer(kind=int16), intent(out) :: oldXsize, oldYsize
    real(kind=single), dimension(:,:,:), pointer ::&
      images
real(single), pointer :: tempImage(:,:)
    real(kind=single), dimension(:,:), allocatable ::&
      netFluxImage
    real(kind=single), dimension(xsize, ysize) ::&
      image, unbinnedNetFluxImage
    integer(kind=int16) :: band
    type(DataSetT) ::&
      inSet
    type(ArrayT) ::&
      imageArray
!    integer, dimension(:), pointer ::&
!      dimVector
    !integer, dimension(numSuppliedBands) ::&	! numSuppliedBands is non-parameter global variable.
    integer, allocatable ::&
      tempXsize(:),&
      tempYsize(:)
!    integer(kind=int8), dimension(:,:), pointer ::&
!      arrayDataInt8
!    integer(kind=int16), dimension(:,:), pointer ::&
!      arrayDataInt16
!    integer(kind=int32), dimension(:,:), pointer ::&
!      arrayDataInt32
!    real(kind=single), dimension(:,:), pointer ::&
!      arrayDataReal32
!    real(kind=double), dimension(:,:), pointer ::&
!      arrayDataReal64
    real(single) ::&
      fracPixelsSaturated,&
      hotPixelCutoff
!, afterMaxFlux

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading image files.')

    allocate(tempXsize(numSuppliedBands), tempYsize(numSuppliedBands))
    oldXsize = xsize
    oldYsize = ysize

    unbinnedNetFluxImage = 0.0
    if (rebinImagePlease) then
      allocate(images(numSuppliedBands, newNumXbins, newNumYbins))
      allocate(netFluxImage(newNumXbins, newNumYbins))

      netFluxImage = 0.0
!      priorMaxFlux = 0.0
!      afterMaxFlux = 0.0
      do band = 1, numSuppliedBands
        inSet = dataSet(imageSetName(band), READ)
        imageArray = array(inSet, 0, READ)

        if (band == 1) call getKeywords(inSet, imageArray)

call readArrayData(imageArray, tempImage) ! in ssclib/dal_aux
!        dimVector => dimensions(imageArray)
!        tempXsize(band) = dimVector(0)
!        tempYsize(band) = dimVector(1)
tempXsize(band) = size(tempImage, 1)
tempYsize(band) = size(tempImage, 2)

        image = 0.0 ! So that, if images are different sizes, the small ones
        ! will get padded with zeroes to the same size as the largest.
image(1:tempXsize(band), 1:tempYsize(band)) = tempImage
deallocate(tempImage)

!        if (dataType(imageArray) == INTEGER8) then
!          arrayDataInt8 => int8Array2Data(imageArray)
!          image(1:tempXsize(band), 1:tempYsize(band))&
!          = real(arrayDataInt8, single)
!        else if (dataType(imageArray) == INTEGER16) then
!          arrayDataInt16 => int16Array2Data(imageArray)
!          image(1:tempXsize(band), 1:tempYsize(band))&
!          = real(arrayDataInt16, single)
!        else if (dataType(imageArray) == INTEGER32) then
!          arrayDataInt32 => int32Array2Data(imageArray)
!          image(1:tempXsize(band), 1:tempYsize(band))&
!          = real(arrayDataInt32, single)
!        else if (dataType(imageArray) == REAL32) then
!          arrayDataReal32 => real32Array2Data(imageArray)
!          image(1:tempXsize(band), 1:tempYsize(band))&
!          = arrayDataReal32
!        else ! this gives a DAL error if data type is none of the above.
!          arrayDataReal64 => real64Array2Data(imageArray)
!          image(1:tempXsize(band), 1:tempYsize(band))&
!          = real(arrayDataReal64, single)
!        end if

        call release(inSet)

        unbinnedNetFluxImage = unbinnedNetFluxImage + image ! note the different x, y limits - we are using the full width of the array now. No reason not to include those zeroes that may lie between tempXsize(band) and xsize, etc.

!        priorMaxFlux = priorMaxFlux + maxval(image)

        call imageRebinner(image, images(band, :, :))

        netFluxImage = netFluxImage + images(band, :, :)

!        afterMaxFlux = afterMaxFlux + maxval(images(band, 1:newNumXbins&
!        , 1:newNumYbins))
      end do

!!      hotPixelCutoff = hiPixelFracToFluxFrac_new(unbinnedNetFluxImage&
!!      , xsize, ysize, numHotPixels / real(xsize) / real(ysize))
!      hotPixelCutoff = hiPixelFracToFluxFrac(unbinnedNetFluxImage&
!      , xsize, ysize)

fracPixelsSaturated = numHotPixels / real(xsize, single) / real(ysize, single)
! numHotPixels is a parameter and thus global.

hotPixelCutoff = valueAtHistoFraction(unbinnedNetFluxImage, 1.0-fracPixelsSaturated) ! in ssclib/array_utils.

!*** NOTE!: there is a facility to supply a mask image to valueAtHistoFraction. It would be nice to move all this gain stuff if possible to _after_ the mask image is read in in colimplot. Then one could make use of it.


      xsize = newNumXbins
      ysize = newNumYbins

!print *, hotPixelCutoff, gainCorrector

      gainCorrector = (maxval(unbinnedNetFluxImage) * hotPixelCutoff)&
      / maxval(netFluxImage)
!      / afterMaxFlux ! This has been introduced largely to correct a problem that occurs when the user has chosen a gain which gives a nice-looking rebinned image, but then finds that the same image when unrebinned is too dark. This occurs because the peak flux value of an unrebinned image, after --nhotpixels hot pixels have been passed over, may be considerably higher than a binned one after it has had the same treatment, so all the flux values shift down in consequence. This makes it hard to guess the correct value of gain to apply. But now the net gain is divided by gainCorrector before being applied to the image.

!print *, oldXsize, oldYsize
!print *, 'max unbinned', maxval(unbinnedNetFluxImage)
!print *, maxloc(unbinnedNetFluxImage)
!print *, newNumXbins, newNumYbins
!print *, 'max binned', maxval(netFluxImage)
!print *, maxloc(netFluxImage)
!print *, ''
!print *, 'hotPixelCutoff', hotPixelCutoff
!print *, 'gainCorrector',  gainCorrector

    else ! don't rebin the images.
      allocate(images(numSuppliedBands, xsize, ysize))
      images = 0.0 ! So that, if images are different sizes, the small ones
      ! will get padded with zeroes to the same size as the largest. Note
      ! that xsize and ysize were set to the largest values among the images.

      do band = 1, numSuppliedBands

!*** rewrite in terms of readArrayData etc??

        inSet = dataSet(imageSetName(band), READ)
        imageArray = array(inSet, 0, READ)

        if (band == 1) call getKeywords(inSet, imageArray)

call readArrayData(imageArray, tempImage) ! in ssclib/dal_aux
!        dimVector => dimensions(imageArray)
!        tempXsize(band) = dimVector(0)
!        tempYsize(band) = dimVector(1)
tempXsize(band) = size(tempImage, 1)
tempYsize(band) = size(tempImage, 2)
images(band, 1:tempXsize(band), 1:tempYsize(band)) = tempImage
deallocate(tempImage)

!        if (dataType(imageArray) == INTEGER8) then
!          arrayDataInt8 => int8Array2Data(imageArray)
!          images(band, 1:tempXsize(band), 1:tempYsize(band))&
!          = real(arrayDataInt8, single)
!        else if (dataType(imageArray) == INTEGER16) then
!          arrayDataInt16 => int16Array2Data(imageArray)
!          images(band, 1:tempXsize(band), 1:tempYsize(band))&
!          = real(arrayDataInt16, single)
!        else if (dataType(imageArray) == INTEGER32) then
!          arrayDataInt32 => int32Array2Data(imageArray)
!          images(band, 1:tempXsize(band), 1:tempYsize(band))&
!          = real(arrayDataInt32, single)
!        else if (dataType(imageArray) == REAL32) then
!          arrayDataReal32 => real32Array2Data(imageArray)
!          images(band, 1:tempXsize(band), 1:tempYsize(band))&
!          = arrayDataReal32
!        else ! this gives a DAL error if data type is none of the above.
!          arrayDataReal64 => real64Array2Data(imageArray)
!          images(band, 1:tempXsize(band), 1:tempYsize(band))&
!          = real(arrayDataReal64, single)
!        end if

        call release(inSet)

        unbinnedNetFluxImage = unbinnedNetFluxImage&
        + images(band, 1:xsize, 1:ysize) ! note the different x, y limits - we are using the full width of the array now. No reason not to include those zeroes that may lie between tempXsize(band) and xsize, etc.
      end do

!!      hotPixelCutoff = hiPixelFracToFluxFrac_new(unbinnedNetFluxImage&
!!      , xsize, ysize, numHotPixels / real(xsize) / real(ysize))
!      hotPixelCutoff = hiPixelFracToFluxFrac(unbinnedNetFluxImage&
!      , xsize, ysize)

fracPixelsSaturated = numHotPixels / real(xsize, single) / real(ysize, single)
! numHotPixels is a parameter and thus global.

hotPixelCutoff = valueAtHistoFraction(unbinnedNetFluxImage, 1.0-fracPixelsSaturated) ! in ssclib/array_utils.

!*** NOTE!: there is a facility to supply a mask image to valueAtHistoFraction. It would be nice to move all this gain stuff if possible to _after_ the mask image is read in in colimplot. Then one could make use of it.


      gainCorrector = hotPixelCutoff
    end if

!print *, 'rc=', gainCorrector

  end subroutine readAndRebinImages

  !---------------------------------------------------------------------
  subroutine getKeywords(inSet, imageArray)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) ::&
      inSet
    type(ArrayT), intent(in) ::&
      imageArray

!    integer(kind=int16) :: status
    integer :: status

    !-------------------------------------------------------------------
    ! Start

    call getStringKwd('INSTRUME', inSet, imageArray, instrumeStr, status)
    call getStringKwd('FILTER',   inSet, imageArray, filterStr,   status)
    call getStringKwd('OBJECT',   inSet, imageArray, objectStr,   status)
    call getStringKwd('OBSERVER', inSet, imageArray, observerStr, status)
    call getStringKwd('DATE-OBS', inSet, imageArray, dateobsStr,  status)
    call getStringKwd('EXPOSURE', inSet, imageArray, exposureStr, status)

  end subroutine getKeywords

  !---------------------------------------------------------------------
  subroutine imageRebinnerSingle(oldImage, newImage)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: oldImage(:,:)
    real(single), intent(out) :: newImage(:,:)

    real(single) ::&
      oldXedges(size(oldImage,1)+1), oldYedges(size(oldImage,2)+1),&
      newXedges(size(newImage,1)+1), newYedges(size(newImage,2)+1)
    integer :: status, i
    integer(int16) :: oldXsize, oldYsize, newXsize, newYsize
!    integer(int16), pointer :: indices(:)

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, '  Rebinning image.')
!print *, 'imageRebinnerSingle: start'

    oldXsize = size(oldImage,1)
    oldYsize = size(oldImage,2)
    newXsize = size(newImage,1)
    newYsize = size(newImage,2)

!    indices => indicize(oldXsize)
!    oldXedges = real(indices, single)
!    deallocate(indices)
oldXedges = (/(real(i, single), i=0,oldXsize)/)
!    indices => indicize(oldYsize)
!    oldYedges = real(indices, single)
!    deallocate(indices)
oldYedges = (/(real(i, single), i=0,oldYsize)/)

!    indices => indicize(newXsize)
!    newXedges = fracVal(1.0, real(oldXsize), (indices - 1) / real(newXsize - 1))
!    deallocate(indices)
newXedges = (/(fracVal(0.0, real(oldXsize), i / real(newXsize))&
 , i=0,newXsize)/)
!    indices => indicize(newYsize)
!    newYedges = fracVal(1.0, real(oldYsize), (indices - 1) / real(newYsize - 1))
!    deallocate(indices)
newYedges = (/(fracVal(0.0, real(oldYsize), i / real(newYsize))&
 , i=0,newYsize)/)

    call rebinImage(oldImage, oldXedges, oldYedges&
      , newImage, newXedges, newYedges, status)
!print *, 'imageRebinnerSingle: rebinImage status =', status

    if (status < 0) then
      write(errstr, '(a, i2)') 'The rebinner returned with the status &
      &value ', status
call error('badRebinStatus')

    else if (status > 0) then
      write(errstr, '(a, i2)') 'The rebinner returned with the status &
      &value ', status
      call warning('doubtfulRebinStatus')
    end if

!print *, 'imageRebinnerSingle: end'

  end subroutine imageRebinnerSingle

  !---------------------------------------------------------------------
  subroutine imageRebinnerBool(oldImage, newImage)
  !---------------------------------------------------------------------

    implicit none

!    logical(bool), intent(in)  :: oldImage(:,:)
!    logical(bool), intent(out) :: newImage(:,:)
    logical, intent(in)  :: oldImage(:,:)
    logical, intent(out) :: newImage(:,:)

    real(single) ::&
      oldXedges(size(oldImage,1)+1), oldYedges(size(oldImage,2)+1),&
      newXedges(size(newImage,1)+1), newYedges(size(newImage,2)+1)
    integer :: status, i
    integer(int16) :: oldXsize, oldYsize, newXsize, newYsize

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, '  Rebinning image.')

    oldXsize = size(oldImage,1)
    oldYsize = size(oldImage,2)
    newXsize = size(newImage,1)
    newYsize = size(newImage,2)

    oldXedges = (/(real(i, single), i=1,oldXsize)/)
    oldYedges = (/(real(i, single), i=1,oldYsize)/)
    newXedges = (/(fracVal(1.0, real(oldXsize), i / real(newXsize - 1))&
     , i=0,newXsize-1)/)

    newYedges = (/(fracVal(1.0, real(oldYsize), i / real(newYsize - 1))&
     , i=0,newYsize-1)/)

    call rebinImage(oldImage, oldXedges, oldYedges&
      , newImage, newXedges, newYedges, status) ! in ssclib/rebinners

    if (status < 0) then
      write(errstr, '(a, i2)') 'The rebinner returned with the status &
      &value ', status
call error('badRebinStatus')

    else if (status > 0) then
      write(errstr, '(a, i2)') 'The rebinner returned with the status &
      &value ', status
      call warning('doubtfulRebinStatus')
    end if

  end subroutine imageRebinnerBool

  !---------------------------------------------------------------------
  subroutine checkNoNegatives(images, xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(inout) :: xsize, ysize
    real(single), dimension(:,:,:), pointer ::&
      images
    integer(int16) :: i

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Checking on negative values.')

    do i = 1, numSuppliedBands
      if (minval(images(i, 1:xsize, 1:ysize)) < 0.0) then
        write(errstr, '(a, i3)') 'Negative image values were found &
        &in image', i
        call warning('negativeValues')
        if (negRemovalStyle == 'offset') then
          images(i, 1:xsize, 1:ysize) = images(i, 1:xsize, 1:ysize)&
          - minval(images(i, 1:xsize, 1:ysize))
        else ! truncate
          where(images(i, 1:xsize, 1:ysize) < 0.0)
            images(i, 1:xsize, 1:ysize) = 0.0
          end where
        end if
      end if
    end do

  end subroutine checkNoNegatives

  !---------------------------------------------------------------------
  subroutine energyInit
  !---------------------------------------------------------------------

    implicit none

    integer(int16) :: i
!, j
!    type(intervalType), pointer :: energyRanges(:)
type(RangeT), pointer :: energyRanges(:)
    logical, allocatable :: noBandInfo(:)

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Calculating energy bands.')

    if (refClrStyle == 'bb') then
      if (userWhiteStyle == 'kev') then
!        whiteMegaKelvin = ((whitekeV * 1000.0) * myElectronQ / boltzmannConst&
whiteMegaKelvin = ((whitekeV * 1000.0) * ElectronQ / Boltzmann&
        / bbPeakConst) / 1.0e6

      else if (userWhiteStyle == 'angstroms') then
        whiteMegaKelvin = (plancksConst * mySpeedOfLight / (whiteAngstroms&
!        * 1.0e-10) / boltzmannConst / bbPeakConst) / 1.0e6
* 1.0e-10) / Boltzmann / bbPeakConst) / 1.0e6

      else if (userWhiteStyle .ne. 'kelvin') then
call error('badValue', 'Unrecognised value of userwhitestyle')
      end if
    end if ! refClrStyle == 'bb'

    allocate(noBandInfo(numSuppliedBands))
    noBandInfo = .false. ! vector default
    do i = 1, numSuppliedBands
      call getEnergyRanges(imageSetName(i), energyRanges, noBandInfo(i))

    if (noBandInfo(i)) exit
      call calcBandCentre(energyRanges, imageBandCentres(i))

      if (refClrStyle == 'bb') then
        bbRefColour(i) = integrateSpectrum(energyRanges, whiteMegaKelvin)
      end if
    end do ! band

    if (any(noBandInfo)) then
      do i = 1, numSuppliedBands
        imageBandCentres(i) = real(i) ! last-resort makeshift!
!print *, 'i, imageBandCentres(i):', i, imageBandCentres(i)
      end do

      if (refClrStyle == 'bb') then
        bbRefColour = 1.0 ! vector op.
      end if
    end if

    averageBandCentre = sum(imageBandCentres) / real(numSuppliedBands)

    deallocate(energyRanges, noBandInfo)
!call error('','')

  end subroutine energyInit

  !---------------------------------------------------------------------
  subroutine getEnergyRanges(inSetName, EeVRanges, noBandInfo)
  !---------------------------------------------------------------------

    implicit none

    character(len=255), intent(in) ::&
      inSetName
    type(RangeT), pointer :: EeVRanges(:)	! intent(in)
    logical, intent(out) :: noBandInfo

    type(RangeT), pointer :: piRanges(:) => null()
    type(DataSetT) ::&
      inSet
    type(BlockT) ::&
      inBlock
    type(ArrayT) ::&
      imageArray
    type(DssT) ::&
      dssPointer
    character(len=20) :: OMfilterName
!    integer(int16) :: numRangesInDSS, maxNumRangesInDSS,&
!      index, rangeNum
!    integer(int16), allocatable :: numRanges(:)
    integer(int32) :: bandNum
!    type(DScompT) :: component
!    type(DSfilterT) :: filter
!    type(intervalsType), allocatable :: allDSSranges(:)
!    type(intervalsType), pointer :: ranges(:)
!    type(intervalType), pointer :: ANDedRanges(:), dummyRanges(:)
!    type(RangeT), dimension(:), pointer ::&
!      energyPIranges
    real(double) :: loEeV, hiEeV
!    integer :: compNum, filtNum, aoStatus
!    integer :: compNum, filtNum, aoStatus
    integer :: status
!integer(int16) :: status

    !-------------------------------------------------------------------
    ! Start:

    noBandInfo = .false. ! default

    inSet = dataSet(inSetName, READ)
    inBlock = block(inSet, 0, READ)
    imageArray = array(inSet, 0, READ)

    if (notUseableInstrument) then
      allocate(EeVRanges(1))
      EeVRanges(1)%lower%type = UNDEFINED
      EeVRanges(1)%upper%type = UNDEFINED
      noBandInfo = .true.
      call release(inSet)
  return
    end if

    if (instrument == EMOS1 .or. instrument == EMOS2&
    .or. instrument == EPN) then
      if (hasDss(imageArray)) then ! in ssclib/dss_aux
        dssPointer = dataSubspace(inBlock)
        call readRanges(dssPointer, 'PI', piRanges) ! in ssclib/dss_aux
if (size(piRanges) <= 0) call error('','')
        call correctRanges(piRanges) ! in ssclib/dss_aux
!        EeVranges => convertPiRangesToEeV(piRanges) ! in ssclib/exposure. The
!        ! call deallocates piRanges.
allocate(EeVRanges(size(piRanges,1)))
EeVranges = convertPiToEeV(piRanges) ! in ssclib/exposure.
!EeVranges => convertPiToEeV(piRanges) ! in ssclib/exposure.
deallocate(piRanges)
        call dssRelease(dssPointer)

      else ! the dataset has no DSS.
!*** warn?

        allocate(EeVRanges(1))
        call getIntKwd('BAND', inSet, imageArray, bandNum, status)
        if (status == 0) then ! energy band info resides in the
        ! BAND keyword:

          EeVRanges(1)%lower%type = INCLUSIVE
          EeVRanges(1)%upper%type = INCLUSIVE
          EeVRanges(1)%lower%value = bandLoEeV(bandNum)
          EeVRanges(1)%upper%value = bandHiEeV(bandNum)
!**** These should be stored centrally - in the cal?

        else
!*** Should I get the actual min and max vals of PI from the appropriate keywords??
          EeVRanges(1)%lower%type = UNDEFINED
          EeVRanges(1)%upper%type = UNDEFINED
          noBandInfo = .true.
!**** warn!!!!!
        end if
      end if ! the dataset has a DSS.
      call release(inSet)

!**** isn't this similar to the routine in lccorr_corr.f90?

    else ! instrument should be OM. Others should be screened out in
    ! subroutine initAndChecks.
      call getStringKwd('FILTER', inSet, imageArray, OMfilterName, status)
if (status .ne. 0) call error('noOMfilter', 'There is no OM filter keyword.')
      OMfilter = filterIdToInt(OMfilterName)
      select case(OMfilter)
      case (UVW2)
        loEeV = angstroms2eV(1800.0d0)
        hiEeV = angstroms2eV(2800.0d0)
      case (UVM2)
        loEeV = angstroms2eV(1800.0d0)
        hiEeV = angstroms2eV(3000.0d0)
      case (UVW1)
        loEeV = angstroms2eV(2200.0d0)
        hiEeV = angstroms2eV(4000.0d0)
      case (U)
        loEeV = angstroms2eV(3000.0d0)
        hiEeV = angstroms2eV(3900.0d0)
      case (B)
        loEeV = angstroms2eV(3700.0d0)
        hiEeV = angstroms2eV(5000.0d0)
      case (V)
        loEeV = angstroms2eV(5000.0d0)
        hiEeV = angstroms2eV(5900.0d0)
      case (WHITE)
        loEeV = angstroms2eV(1600.0d0)
        hiEeV = angstroms2eV(7000.0d0)
      case default
call error('badOMfilter', "Can't calculate a bandpass for this OM &
&filter setting. Use refclrstyle=auto.")
      end select

      allocate(EeVRanges(1))
      EeVRanges(1)%lower%value = loEeV
      EeVRanges(1)%upper%value = hiEeV
    end if

  end subroutine getEnergyRanges

  !---------------------------------------------------------------------
  subroutine getNetFluxImage(images, xsize, ysize, netFluxImage)
  !---------------------------------------------------------------------

    implicit none

    real(kind=single), dimension(:,:,:), pointer :: images
    integer(kind=int16), intent(in) :: xsize, ysize
    real(kind=single), dimension(xsize, ysize), intent(out) :: netFluxImage
    integer(kind=int16) :: band

    !-------------------------------------------------------------------
    ! Start:

    netFluxImage = 0.0 ! array op.
    do band = 1, numSuppliedBands
      netFluxImage = netFluxImage + images(band, 1:xsize, 1:ysize) ! array op.
    end do
    netFluxImage = netFluxImage / maxval(netFluxImage) ! It is important
    ! to normalise netFluxImage because it is compared to
    ! meanCalcRelFluxCutoff in divideByRefColour.

  end subroutine getNetFluxImage

  !---------------------------------------------------------------------
  subroutine loadAndRebinMask(mask, xsize, ysize, oldXsize, oldYsize)
  !---------------------------------------------------------------------

    implicit none

!    type(DataSetT) :: inSet
!    type(ArrayT) :: imageArray
!    integer, dimension(:), pointer ::&
!      dimVector
!    integer(kind=int8), dimension(:,:), pointer ::&
!      arrayDataInt8
!    integer(kind=int16), dimension(:,:), pointer ::&
!      arrayDataInt16
!    integer(kind=int32), dimension(:,:), pointer ::&
!      arrayDataInt32
!    real(kind=single), dimension(:,:), pointer ::&
!      arrayDataReal32
!    real(kind=double), dimension(:,:), pointer ::&
!      arrayDataReal64

!    logical(bool), pointer :: mask(:,:)
    logical,        pointer    :: mask(:,:)
    integer(int16), intent(in) :: xsize, ysize, oldXsize, oldYsize

    integer :: tempXsize, tempYsize
!    logical(bool), dimension(:,:), pointer ::&
    logical, dimension(:,:), pointer ::&
      tempMask
!    real(kind=single), dimension(:,:), allocatable ::&
!      image0, image1

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading detector mask.')

allocate(mask(xsize, ysize))

call readArrayData(maskSetName, tempMask) ! in ssclib/dal_aux
tempXsize = size(tempMask, 1)
tempYsize = size(tempMask, 2)

!    inSet = dataSet(maskSetName, READ)
!    imageArray = array(inSet, 0, READ)
!    dimVector => dimensions(imageArray)
!    tempXsize = dimVector(0)
!    tempYsize = dimVector(1)
    if (tempXsize .ne. oldXsize .or. tempYsize .ne. oldYsize) then
      call warning('maskDifferentSize', 'The mask image is not the same size &
      &as the largest of the input images. It will be rebinned to the same &
      &size.')
    end if

    if (rebinImagePlease .or. tempXsize .ne. oldXsize .or. tempYsize .ne. oldYsize)&
    then
mask = .false.
call imageRebinner(tempMask, mask)

!      allocate(image0(tempXsize, tempYsize))
!      allocate(image1(xsize, ysize))
!
!      if (dataType(imageArray) == INTEGER8) then
!        arrayDataInt8 => int8Array2Data(imageArray)
!        image0(1:tempXsize, 1:tempYsize)&
!        = real(arrayDataInt8, single)
!      else if (dataType(imageArray) == INTEGER16) then
!        arrayDataInt16 => int16Array2Data(imageArray)
!        image0(1:tempXsize, 1:tempYsize)&
!        = real(arrayDataInt16, single)
!      else if (dataType(imageArray) == INTEGER32) then
!        arrayDataInt32 => int32Array2Data(imageArray)
!        image0(1:tempXsize, 1:tempYsize)&
!        = real(arrayDataInt32, single)
!      else if (dataType(imageArray) == REAL32) then
!        arrayDataReal32 => real32Array2Data(imageArray)
!        image0(1:tempXsize, 1:tempYsize)&
!        = arrayDataReal32
!      else ! this gives a DAL error if data type is none of the above.
!        arrayDataReal64 => real64Array2Data(imageArray)
!        image0(1:tempXsize, 1:tempYsize)&
!        = real(arrayDataReal64, single)
!      end if
!
!      call release(inSet)
!
!      call imageRebinner(image0, image1)
!      deallocate(image0)

    else ! don't rebin the mask image.
mask = tempMask
!      allocate(image1(xsize, ysize))
!
!      if (dataType(imageArray) == INTEGER8) then
!        arrayDataInt8 => int8Array2Data(imageArray)
!        image1(1:xsize, 1:ysize)&
!        = real(arrayDataInt8, single)
!      else if (dataType(imageArray) == INTEGER16) then
!        arrayDataInt16 => int16Array2Data(imageArray)
!        image1(1:xsize, 1:ysize)&
!        = real(arrayDataInt16, single)
!      else if (dataType(imageArray) == INTEGER32) then
!        arrayDataInt32 => int32Array2Data(imageArray)
!        image1(1:xsize, 1:ysize)&
!        = real(arrayDataInt32, single)
!      else if (dataType(imageArray) == REAL32) then
!        arrayDataReal32 => real32Array2Data(imageArray)
!        image1(1:xsize, 1:ysize)&
!        = arrayDataReal32
!      else ! this gives a DAL error if data type is none of the above.
!        arrayDataReal64 => real64Array2Data(imageArray)
!        image1(1:xsize, 1:ysize)&
!        = real(arrayDataReal64, single)
!      end if
!
!      call release(inSet)

    end if

!    allocate(mask(xsize, ysize))
!    mask = .false.
!    where(image1 > 0.0) mask = .true.
deallocate(tempMask)
!call release(inSet)

  end subroutine loadAndRebinMask

  !---------------------------------------------------------------------
!  subroutine calcGain(netFluxImage, mask, xsize, ysize)
subroutine calcGain(netFluxImage, mask)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) ::&
      netFluxImage(:,:)
!    logical(bool), intent(in) ::&
    logical, intent(in) ::&
      mask(size(netFluxImage,1),size(netFluxImage,2))

    real(single) ::&
      presentFluxFrac
    integer(int16) ::&
      xsize,&
      ysize

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Calculating image gain.')
!print *, 'start of calcGain'

    xsize = size(netFluxImage,1)
    ysize = size(netFluxImage,2)

!!    presentFluxFrac = loPixelFracToFluxFrac(netFluxImage, mask, xsize, ysize&
!presentFluxFrac = loPixelFracToFluxFrac(mask, netFluxImage, kPixelFrac)

presentFluxFrac = valueAtHistoFraction(netFluxImage, kPixelFrac, mask)
!presentFluxFrac = valueAtHistoFraction(netFluxImage, kPixelFrac, logical(mask))
!****** mask arg to valueAtHistoFraction is logical not logical(bool) since ssclib-3.17.
! kPixelFrac is a command-line parameter and thus global. The function valueAtHistoFraction is in ssclib/array_utils.

    if (presentFluxFrac .le. 0.0) then ! warn, and leave the gain alone.
      call warning('infAutoGain', "Auto gain calculation failed because &
      &there are too many zero-valued pixels in the image. Falling &
      &back to 'user' value of gain.")

    else
      gain = kFluxFrac / presentFluxFrac * gainCorrector

      write(errstr, '(a, es11.4)') '  Gain = ', gain
!call error('TestTest', 'Here00.')
      call message(AppMsg, Verbose)
    end if

!print *, 'got to end of calcGain'

  end subroutine calcGain

  !---------------------------------------------------------------------
!  subroutine divideByRefColour(images, netFluxImage, mask, xsize, ysize)
subroutine divideByRefColour(images, netFluxImage, mask)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer    :: images(:,:,:)
    real(single), intent(in) :: netFluxImage(:,:)
!    logical(bool), intent(in) :: mask(size(netFluxImage,1),size(netFluxImage,2))
    logical,      intent(in) :: mask(size(netFluxImage,1),size(netFluxImage,2))

    real(single), allocatable :: refColour(:)
    integer(int16) :: xsize, ysize
    integer(int16) :: i, xi, yi
    character(len=64) :: s

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Calculating reference colour.')

    xsize = size(netFluxImage,1)
    ysize = size(netFluxImage,2)

    allocate(refColour(numSuppliedBands))

    if (refClrStyle == 'internal') then
      if (autoWhiteStyle == 'rms') then
        refColour = 0.0 ! vector op
        do xi = 1, xsize
          do yi = 1, ysize
            if (mask(xi, yi) .and. netFluxImage(xi, yi)&
            > meanCalcRelFluxCutoff) then
              refColour = refColour + images(1:numSuppliedBands, xi, yi)&
              * images(1:numSuppliedBands, xi, yi)
            end if
          end do ! yi
        end do ! xi
        refColour = sqrt(refColour) / real(xsize) / real(ysize)

      else if (autoWhiteStyle == 'mean') then
        refColour = 0.0 ! vector op
        do xi = 1, xsize
          do yi = 1, ysize
            if (mask(xi, yi) .and. netFluxImage(xi, yi)&
            > meanCalcRelFluxCutoff) then
              refColour = refColour + images(1:numSuppliedBands, xi, yi)
            end if
          end do ! yi
        end do ! xi
        refColour = refColour / real(xsize) / real(ysize)

      else
call error('badValue', 'Unrecognised value of autowhitestyle')
      end if ! autoWhiteStyle

    else if (refClrStyle == 'bb') then
      refColour = bbRefColour ! vector op

    else if (refClrStyle == 'user') then
      refColour = userRefColour ! vector op

    else if (refClrStyle == 'none') then
      refColour = 1.0 ! vector op

    else
call error('badValue', 'Unrecognised value of refclrstyle')
    end if ! refClrStyle

    do i = 1, numSuppliedBands
      if (refColour(i) .le. 0.0) then
call error('zeroWhiteFlux', "One of the values of the reference &
&colour equals zero.")
      end if
    end do

    refColour = refColour / maxval(refColour) ! vector op

    do i = 1, numSuppliedBands
      images(i, 1:xsize, 1:ysize)&
      = images(i, 1:xsize, 1:ysize) / refColour(i)
    end do

    if (verbosity(AppMsg) >= Verbose) then
      if ( numSuppliedBands == 3 ) then
        write(errstr, '(a, f10.5, f10.5, f10.5)') 'Reference RGB =', refColour(red), refColour(grn), refColour(blu)
      else
        write(errstr, '(a)') 'Reference Colours ='
        do i = 1, numSuppliedBands
          write(s, '(f10.5)') refColour(i)
          errstr = trim(errstr) // s
        end do
      end if
      call message(AppMsg, Verbose)
!print *, refColour
    end if
    deallocate(refColour)

  end subroutine divideByRefColour

  !---------------------------------------------------------------------
  subroutine convertToRGB(purplicity, thermalClr, rgb, status)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: purplicity, thermalClr
    real(single), intent(out) :: rgb(red:blu)
    integer(int16), intent(out) :: status

    integer(int16) :: clr
    real(single) :: x, y

    !-------------------------------------------------------------------
    ! Start

    if (purplicity >= 0.0) then
      if (thermalClr < 1.0) then		! A
        rgb(red) = 1.0
        rgb(grn) = thermalClr * (1.0 - purplicity)
        rgb(blu) = thermalClr * purplicity
      else if (thermalClr < 2.0) then		! B1?
        y = 3.0 - thermalClr
        rgb(blu) = thermalClr - 1.0 + y * purplicity / 2.0
        if (rgb(blu) > 1.0) then			! no, B2
          rgb(red) = y * (1.0 - purplicity / 2.0)
!*** check for rgb(red) .le. 1.0? (bug error if not)
          rgb(grn) = y * (1.0 - purplicity)
          rgb(blu) = 1.0
        else					! yes, B1
          rgb(red) = 1.0
          rgb(grn) = 1.0 - y * purplicity / 2.0
        end if
      else					! B2
        y = 3.0 - thermalClr
        rgb(red) = y * (1.0 - purplicity / 2.0)
        rgb(grn) = y * (1.0 - purplicity)
        rgb(blu) = 1.0
      end if
    else ! purplicity < 0:
      if (thermalClr < 1.5) then		! E
      ! Note: have already screened out (a few statements ago) cases
      ! in which purplicity < 0 and thermalClr < 1. Such combinations
      ! can occur if a point with thermalClr initially a little
      ! greater than 1 are distorted by a negative 'heat' value.
      ! These new points have nowhere to go on the hue diagram and
      ! so the negative purplicity is truncated to zero, in which
      ! case they will get caught by the test 'if (purplicity >=
      ! 0.0)' above and never reach here..
        x = thermalClr - 1.0
        rgb(red) = 1.0 + 2.0 * x * purplicity
        rgb(grn) = 1.0
        rgb(blu) = x * (1.0 + purplicity)
      else if (thermalClr < 2.0) then		! D2
        rgb(red) = 1.0 + purplicity
        rgb(grn) = 1.0
        rgb(blu) = thermalClr - 1.0 + purplicity / 2.0
      else if (thermalClr < 2.5) then		! D1?
        rgb(grn) = 3.0 - thermalClr - purplicity / 2.0
        if (rgb(grn) > 1.0) then			! no, D2
          rgb(red) = 1.0 + purplicity
          rgb(grn) = 1.0
          rgb(blu) = thermalClr - 1.0 + purplicity / 2.0
!*** check for rgb(blu) .le. 1.0? (bug error if not)
        else					! yes, D1
          rgb(red) = 3.0 - thermalClr + purplicity / 2.0
          rgb(blu) = 1.0
        end if
      else					! C
        y = 3.0 - thermalClr
        rgb(red) = y * (1.0 + purplicity)
        rgb(grn) = y * (1.0 - purplicity)
        rgb(blu) = 1.0
      end if
    end if

    do clr = red, blu
      if (rgb(clr) > 1.0 .or. rgb(clr) < 0.0) then
        if (rgb(clr) > 1.0001 .or. rgb(clr) < -0.0001) then
          status = -1
  return
        end if

        rgb(clr) = min(1.0, max(0.0, rgb(clr)))
      end if
    end do

  end subroutine convertToRGB

  !---------------------------------------------------------------------
!  subroutine netFluxTransform(netFluxImage, mask, xsize, ysize)
subroutine netFluxTransform(netFluxImage, mask)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: netFluxImage(:,:)
!    logical(bool), intent(in) :: mask(size(netFluxImage,1),size(netFluxImage,2))
    logical,      intent(in) :: mask(size(netFluxImage,1),size(netFluxImage,2))

    integer(int16) :: xsize, ysize

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Transforming net flux.')

    xsize = size(netFluxImage, 1)
    ysize = size(netFluxImage, 2)

if (maxval(netFluxImage) .le. 0.0) call error('badNetFluxImage','')
    if (minval(netFluxImage) < 0.0) then
      call warning('suspectNetFluxImage', 'Some <0 fluxes detected. These &
      &will be truncated to zero.')

      where (netFluxImage < 0.0) netFluxImage = 0.0
    end if

    if (fluxScalingStyle == 'equ') then
      call equalWeight(netFluxImage, xsize, ysize)
    else if (fluxScalingStyle == 'log') then
      call logscale(netFluxImage, xsize, ysize)
    end if

    where (.not. mask) netFluxImage = 0.0 ! probably unnecessary here since
    ! all non-mask values are set to 1.0 in finalColourTweak.

  end subroutine netFluxTransform

  !---------------------------------------------------------------------
  subroutine equalWeight(netFluxImage, xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) :: xsize, ysize
    real(kind=single), intent(inout) :: netFluxImage(xsize, ysize)

    integer(kind=int16), parameter :: numBins = 1000
    integer(kind=int32) :: weights(numBins), accumWeight(0:numBins)
    integer(kind=int16) :: xi, yi, bin
    real(kind=single) :: maxFlux, maxFluxPlus, binFrac
    integer(kind=int32) :: index

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Transform of total flux scale: &
    &equal weight.')

    maxFlux = maxval(netFluxImage)
    maxFluxPlus = nearest(maxFlux, 1.0) ! Just to prevent bin >= numBins
    weights = 0

    do xi = 1, xsize
      do yi = 1, ysize
        if (netFluxImage(xi, yi) > 0.0) then
          bin = 1 + floor((netFluxImage(xi, yi) / maxFluxPlus) * numBins)
          weights(bin) = weights(bin) + 1.0
        end if
      end do
    end do
    accumWeight(0) = 0
    do bin = 1, numBins
      accumWeight(bin) = accumWeight(bin - 1) + weights(bin)
    end do

    do xi = 1, xsize
      do yi = 1, ysize
        index = 1 + xi + xsize * yi
        if (netFluxImage(xi, yi) > 0.0) then
          bin = 1 + floor((netFluxImage(xi, yi) / maxFluxPlus) * numBins)

          binFrac = (netFluxImage(xi, yi) / maxFlux) * numBins - (bin - 1)
          netFluxImage(xi, yi) = accumWeight(bin - 1) + binFrac * weights(bin)
        end if
      end do
    end do

!****** check that this sub leaves netFluxImage normalised!!! (ie, with values between 0 and 1)

  end subroutine equalWeight

  !---------------------------------------------------------------------
  subroutine logscale(netFluxImage, xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) :: xsize, ysize
    real(kind=single), intent(inout) :: netFluxImage(xsize, ysize)

    integer(kind=int16) :: xi, yi
    real(kind=single) :: minNonZero, offset

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Transform of total flux scale: &
    &log scale.')

    ! Find first value of the array that is > 0:
    XPIXEL: do xi = 1, xsize
      do yi = 1, ysize
        if (netFluxImage(xi, yi) > 0.0) then
          minNonZero = netFluxImage(xi, yi)
    exit XPIXEL
        end if ! netFluxImage should be checked in the calling routine
        ! so that this procedure is guaranteed to find at least one >0 pixel.
      end do
    end do XPIXEL

    ! Now find the smallest positive-valued member of the array:
    do xi = 1, xsize
      do yi = 1, ysize
        if (netFluxImage(xi, yi) > 0.0&
        .and. netFluxImage(xi, yi) < minNonZero) then
          minNonZero = netFluxImage(xi, yi)
        end if
      end do
    end do

    offset = log(minNonZero)
    where (netFluxImage >= minNonZero)
      netFluxImage = nearest((log(netFluxImage) - offset), 1.0)
    end where
    ! The above is meant to ensure that (i) all zero values of the input
    ! array stay at zero; (ii) all >0 values of the input array end up
    ! still > 0; (iii) there is not a big gap between 0 and the start of
    ! range of output values.

    ! Now add some offset so the minimum value of the resulting array
    ! divided by the max val is equal to the parameter :
    netFluxImage = netFluxImage + maxval(netFluxImage) * bkgLevelFrac&
    / (1.0 - bkgLevelFrac)

!****** check that this sub leaves netFluxImage normalised!!! (ie, with values between 0 and 1)

  end subroutine logscale

  !---------------------------------------------------------------------
  subroutine recombineImageInfo(rgbImage, netFluxImage, xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) :: xsize, ysize
    real(kind=single), intent(inout) :: netFluxImage(xsize, ysize),&
      rgbImage(xsize, ysize, red:blu)
    real(kind=single) :: maxNetFlux

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Recombining colour and flux values.')

    maxNetFlux = maxval(netFluxImage)
    netFluxImage = netFluxImage / maxNetFlux ! scales net flux to between
    ! 0 and 1.
    rgbImage(1:xsize, 1:ysize, red) = rgbImage(1:xsize, 1:ysize, red)&
    * netFluxImage
    rgbImage(1:xsize, 1:ysize, grn) = rgbImage(1:xsize, 1:ysize, grn)&
    * netFluxImage
    rgbImage(1:xsize, 1:ysize, blu) = rgbImage(1:xsize, 1:ysize, blu)&
    * netFluxImage

  end subroutine recombineImageInfo

  !---------------------------------------------------------------------
  subroutine netFluxSaturate(rgbImage, netFluxImage, xsize, ysize)
  !---------------------------------------------------------------------
  ! At this point the colour value information is contained in the array
  ! rgbImage. The values in the array are reals in the range 0 to 1. The
  ! relative net flux for a given pixel (absolute values don't matter and
  ! no attempt has been made in the program to preserve them) is obtainable
  ! by averaging the three rgbImage values of that pixel but it is
  ! convenient to keep the previously calculated array netFluxImage, which
  ! already contains these values. (Alternatively, one can find the component
  ! of the rgb coordinates along the vector (1,1,1) and divide this by
  ! root(3). Mathematically speaking there is no difference, but the latter
  ! may assist in visualising the procedure in this subroutine.)
  !
  ! Shortly in the program the values in rgbImage will be translated directly
  ! to the output image. However it may be that most of the space in the rgb
  ! 'cube' is occupied by a relatively small number of pixels. This will
  ! result in the image being dark except for a few bright pixels. This can
  ! occur because the original image came from a ccd dominated by a small
  ! number of defective 'hot' pixels, or it may simply be that the image
  ! contains one or more very bright sources. In either case, the brightest
  ! pixel (after the colour manipulations that have so far been carried out)
  ! has been used up til now to scale the net flux values of all the pixels
  ! and thus may have pushed the bulk of them down into the 'black' corner.
  ! It is often desirable to rescale the net flux values so that the bulk of
  ! the pixels have rgb coordinates about in the middle of the cube. The
  ! bright pixels will thus become 'saturated', ie pushed into the 'white'
  ! corner of the rgb colour cube.
  !
  ! This saturation procedure is carried out, for each pixel, in two steps,
  ! described below.
  !
  ! STEP 1: The purpose of this first step is to choose a new value of net
  ! flux for each pixel, such that the small values become amplified and
  ! the values close to maxNetFlux become even closer. This is done by use
  ! of the hyperbolic function in the subroutine bias().
  !
  ! There are two inputs that are used to decide by how much to 'stretch'
  ! the net flux values: the number of too-hot pixels specified by the
  ! command-line parameter numHotPixels, and the saturation fraction
  ! specified by the command-line parameter satAtFluxFrac.
!******** now inverted and called 'gain'.
  !                                                        The saturation
  ! fractions associated with these two parameters are essentially multiplied
  ! together. Eg suppose numHotPixels = 6 and satAtFluxFrac = 0.7. The
  ! function hiPixelFracToFluxFrac_new()
!******** this call has now been moved to the rebinning sub.
  !                                 (in effect) calculates the flux of the
  ! dimmest of the 6 brightest pixels, and the fraction of the maximum net
  ! flux that corresponds to. Suppose that this fraction turned out to be
  ! 0.46. Then the 'stretch' is calculated as 1 / (0.7 * 0.46) = 3.1. The
  ! old values of net flux are then suitably scaled and passed through the
  ! function bias() with a value of bend = (1.0 - stretch) / (1.0 + stretch)
  ! = -0.512. This multiplies all the very small values of net flux by 3.1.
  !
  ! Note that, because of the transformations of the net flux and the colour
  ! values that have already been carried out in the program, the order of
  ! decreasing brightness of pixels may now be different than for the raw
  ! images.
  !
  ! STEP 2: The purpose of this step is to calculate, for each pixel, a new
  ! triplet of rgb values to correspond to the new net flux. This is done
  ! by forcing the rgb values for a given pixel to follow a (by coincidence)
  ! hyperbolic path within the colour cube. For any given rbg point, a unique
  ! path may be found connecting that point with the (0,0,0) and (1,1,1)
  ! points. At each location on this path, the net flux is well defined as
  ! the component of that location in the (1,1,1) direction,  divided by
  ! root(3). It follows that the reverse is true, ie that any given net flux
  ! is associated with a given location on the path. One can think of the
  ! process as sliding a bead along a hyperbola-shaped wire within the
  ! colour cube. (In fact I cheat a bit and the 'bead' is not slid to a
  ! position corresponding exactly with the new net flux, but close enough.)
  !
  ! The first problem is to define the shape of the 'wire' for each pixel.
  ! One can start by specifying that the curve should be such that the hue
  ! of the pixel should be the same for any point on the curve. Obviously
  ! we do not want flux manipulations to affect the hue. This can be
  ! visualised by once again rotating the colour cube until one is viewing
  ! it back along the (1,1,1) axis in the direction of the origin. Points
  ! having the same hue occupy planes within the cube that all pass through
  ! the (1,1,1) vector and are hence all seen edge-on from this direction.
  ! Thus one can reduce by 1 the number of degrees of freedom of the curve
  ! by dissecting the colour cube with a plane that intersects both the
  ! (1,1,1) axis and the (old) rgb coordinates of the pixel. The 'wire'
  ! (and hence necessarily the new rgb coordinates) must therefore lie in
  ! this hue-invariant plane. In addition, we can say further that we are
  ! really only interested in the half-plane that contains the old rgb
  ! coordinates, since to cross the (1,1,1) axis would imply a reversal of
  ! the hue.
  !
  ! Now, it can be shown that if a cube is cut in such a manner, the half-
  ! plane of the cut face can be mapped to the triangle (0,0)-(0,1)-(1,1)
  ! by a linear transformation. The transformation is particularly easy to
  ! arrive at, since the value of the largest of the triplet of rgb values
  ! is equal to the x coordinate in the triangle and the value of the smallest
  ! rgb value is equal to the y coordinate. The curve can thus easily be
  ! found as the hyperbola that connects the three points (0,0), (1,1) and
  ! (largest rgb, smallest rgb) within the triangle coordinate system. Now,
  ! the net flux value is approximately equal to the distance (in triangle
  ! coordinate system) in the direction of the hypotenuse divided by root(2).
  ! So given the new net flux value, it is relatively easy to calculate the
  ! new rgb triplet. This calculation is performed by the subroutine
  ! moveBead().

    implicit none

    integer(kind=int16), intent(in) ::&
      xsize,&
      ysize
    real(kind=single), intent(inout) ::&
      rgbImage(xsize, ysize, red:blu), netFluxImage(xsize, ysize)

    real(single) :: stretch, warpp
    integer(kind=int16) :: xi, yi

!integer :: statusPG, pgopen
!real, dimension(7) :: polyx, polyy

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Calculating flux saturation.')

    stretch = gain * clrGainDistortion / gainCorrector
    ! Note: can have stretch less than 1 if gainCorrector is large.
!print *, 'gain, gainCorrector, stretch:', gain, gainCorrector, stretch

    warpp = (stretch - 1.0) / (stretch + 1.0)
!***** Taylor series if stretch very large or very small??
!***** some explanation of what is going on here.

if (gTestFlag) then
print *, 'gain, clrGainDistortion, gainCorrector, stretch, warp:'
print *, gain, clrGainDistortion, gainCorrector, stretch, warpp
end if
    do xi = 1, xsize
      do yi = 1, ysize
if (xi == sxi .and. yi == syi .and. gTestFlag) then
print *, netFluxImage(xi, yi)
print *, rgbImage(xi, yi, :)
end if
        call netFluxSaturateSlave(rgbImage(xi, yi, :)&
        , netFluxImage(xi, yi), warpp)
if (xi == sxi .and. yi == syi .and. gTestFlag) then
print *, netFluxImage(xi, yi)
print *, rgbImage(xi, yi, :)
end if
      end do
    end do

if (index(testType, 'satcheck') == 1) then
!  scale = 2.0 * root2 / root3
!  call pgmove(real(polyx(7) + y2 / scale), real(polyy(7) + z2 / scale))
!  call pgdraw(real(polyx(7) + newy2 / scale), real(polyy(7) + newz2 / scale))
!
!!  call pgmove(0.1, 0.05)
!!  call pgdraw(0.2, 0.1)
  call pgclos
end if

  end subroutine netFluxSaturate

  !---------------------------------------------------------------------
  subroutine netFluxSaturateSlave(rgb, netFlux, warpp)
  !---------------------------------------------------------------------

    implicit none

    real(kind=single), intent(in) ::&
      warpp
    real(kind=single), intent(inout) ::&
      rgb(red:blu), netFlux

!    integer(kind=int16):: i_min, i_med, i_max
!integer:: i_min, i_med, i_max
integer(int16):: i_max
integer:: i_min, i_med
    integer(kind=int8) :: status, clr
    real(kind=single) :: newNetFluxFrac, oldNetFluxFrac, y2, z2, newy2, newz2

logical :: flag
real :: oldX, oldY, newX, newY, root3

    !-------------------------------------------------------------------
    ! Start:

root3 = RootThree
!******** change all the root3s to RootThree when sure you can link ok with constants module.

flag = .false.

if (flag) then
  print *, '      A', rgb(red), rgb(grn), rgb(blu)
end if

!****** call sort(rgb, indices)

!        i_max = getIndexOfMax(rgb, int(3, int16))
i_max = maxLoc1d(rgb) ! in ssclib/array_utils
        ! Guess:
        i_min = red + modulo(int(i_max), 3)
        i_med = red + modulo(int(i_max) + 1, 3)
!print *, '      i_min, i_med, i_max:', i_min, i_med, i_max

        if (rgb(i_min) > rgb(i_med)) then
        ! guessed wrong.
          i_med = red + modulo(int(i_max), 3)
          i_min = red + modulo(int(i_max) + 1, 3)
!print *, '      oops, should be:', i_min, i_med, i_max
        else
!print *, i_min, i_med, i_max
        end if

if (i_min == i_med .or. i_min == i_max .or. i_med == i_max) &
call error('satBug0','satBug0')

oldNetFluxFrac = netFlux

oldX = real(rgb(i_max))
oldY = real(rgb(i_min))
!***y2 = (rgb(grn) - rgb(blu)) / root2
!***z2 = (rgb(red) * root2 - (rgb(grn) + rgb(blu)) / root2) / root3
y2 = (rgb(grn) - rgb(blu)) / RootTwo
z2 = (rgb(red) * RootTwo - (rgb(grn) + rgb(blu)) / RootTwo) / RootThree

if (flag) then
  print *, '      y2, z2 =', y2, z2
end if

! These numbers (x2, y2, z2) are the coordinates of rgb in a rotated coordinate system. Starting with the colour cube aligned with axes x0 = rgb(blu), y0 = rgb(grn) and z0 = rgb(red), rotate it first by 45 degrees about the z0 axis. The relevant transform is
!	x1 = x0 / root2 + y0 / root2
!	y1 = y0 / root2 - x0 / root2
!	z1 = z0
! Then rotate by asin(1/root3) towards the x1 axis, giving
!	x2 = x1 * root2 / root3 + z1 / root3
!	y2 = y1
!	z2 = z1 * root2 / root3 - x1 / root3
! The full expressions for y2 and z2 are as given above, whereas
!	x2 = (rgb(red) + rgb(grn) + rgb(blu)) / root3
! Thus y2 and z2 are the coordinates of a projected point on the hexagonal hue diagram.

if (flag) then
  print *, '      warp =', warpp
  print *, '      B', rgb(red), rgb(grn), rgb(blu)
end if

        call moveBead(warpp, rgb(i_max), rgb(i_min), status)
if (status .ne. 0) then
  print *, 'non-zero moveBead exit status. Dump:'
  print *, 'i_min, i_med, i_max', i_min, i_med, i_max
  print *, 'rgb(red), rgb(grn), rgb(blu)', rgb(red), rgb(grn), rgb(blu)
  call error('','')
end if
newX = real(rgb(i_max))
newY = real(rgb(i_min))

        if (z2 == 0.0 .and. y2 == 0.0) then
          rgb(i_med) = rgb(i_max) ! all the rbg values should be the same in this case.
        else
          ! Note that the denominators in the following expressions ought never to be equal to zero. Bug if so.
          select case(i_max)
          case(red)
            if (i_min == grn) then
!***              rgb(blu) = ((z2 * root3 + y2) * rgb(grn) - 2.0 * y2 * rgb(red))&
!***              / (z2 * root3 - y2)
rgb(blu) = ((z2 * RootThree + y2) * rgb(grn) - 2.0 * y2 * rgb(red))&
/ (z2 * RootThree - y2)

            else if (i_min == blu) then
!***              rgb(grn) = ((z2 * root3 - y2) * rgb(blu) + 2.0 * y2 * rgb(red))&
!***              / (z2 * root3 + y2)
rgb(grn) = ((z2 * RootThree - y2) * rgb(blu) + 2.0 * y2 * rgb(red))&
/ (z2 * RootThree + y2)
            end if

          case(grn)
            if (i_min == red) then
              rgb(blu) = ((z2 * root3 + y2) * rgb(grn) - 2.0 * y2 * rgb(red))&
              / (z2 * root3 - y2)

            else if (i_min == blu) then
              rgb(red) = ((z2 * root3 + y2) * rgb(grn) - (z2 * root3 - y2)&
              * rgb(blu)) / 2.0 / y2
            end if

          case(blu)
            if (i_min == red) then
              rgb(grn) = ((z2 * root3 - y2) * rgb(blu) + 2.0 * y2 * rgb(red))&
              / (z2 * root3 + y2)

            else if (i_min == grn) then
              rgb(red) = ((z2 * root3 + y2) * rgb(grn) - (z2 * root3 - y2)&
              * rgb(blu)) / 2.0 / y2
            end if
          end select
        end if

        do clr = red, blu
          if (rgb(clr) > 1.0 .or. rgb(clr) < 0.0) then
            if (rgb(clr) > 1.001 .or. rgb(clr) < -0.001) then
print *, clr, i_min, i_med, i_max, rgb(clr)
call error('bugSatRGB0','bugSatRGB0')
            end if

            rgb(clr) = min(1.0, max(0.0, rgb(clr)))
          end if
        end do

!***newy2 = (rgb(grn) - rgb(blu)) / root2
!***newz2 = (rgb(red) * root2 - (rgb(grn) + rgb(blu)) / root2) / root3
newy2 = (rgb(grn) - rgb(blu)) / RootTwo
newz2 = (rgb(red) * RootTwo - (rgb(grn) + rgb(blu)) / RootTwo) / RootThree
!print *, 'new', rgb(red), rgb(grn), rgb(blu), ' ', newy2, newz2, ' ', newz2/newy2
!print *, ''

newNetFluxFrac = (rgb(red) + rgb(grn) + rgb(blu)) / 3.0
netFlux = newNetFluxFrac

if (index(testType, 'satcheck') == 1) then
  call pgmove(oldX, oldY)
  call pgdraw(newX, newY)
!print *, oldNetFluxFrac, newNetFluxFrac, oldX, newX, oldY, newY
end if

  end subroutine netFluxSaturateSlave

  !---------------------------------------------------------------------
  subroutine moveBead(fbend, x, y, status)
  !---------------------------------------------------------------------
  ! In this subroutine we are dealing with a hyperbola described in two
  ! different coordinate systems. In the x-y coordinate system, the equation
  ! of the hyperbola is
  !
  !	         (1 + bend) * x
  !	y = -------------------------.					(1)
  !	     1 - bend + 2 * bend * x
  !
  ! The hyperbola can thus be seen to have the following properties: (i) it
  ! passes through the points (0,0) amd (1,1); (ii) sensible values of 'bend'
  ! range from -1 to +1; for bend = -1, y = 0, but with a singularity at
  ! x = 1; for bend = 0, the equation reduces to the straight line y = x;
  ! for bend = +1, y = 1, with a singularity at x = 0.
  !
  ! The xx-yy system is at 45 degrees to the x-y one and is defined as follows:
  !
  !	x = xx - yy
  !	y = xx + yy.
  !
  ! The equation for the hyperbola in this system is necessarily quadratic, viz:
  !
  !	      1 +- sqrt(1 + 4 * bend^2 * xx * (xx - 1))
  !	yy = ------------------------------------------.		(2)
  !	                    2 * bend
  !
  ! Only values of bend <= 0 are permitted in the subroutine, corresponding
  ! with values of yy <= 0. For this condition, only the '-' of the '+-' is
  ! necessary.
  !
  ! Note there is a singularity in this formula at bend = 0. For bend values
  ! close to that some sort of approximation should be used. I haven't yet
  ! figured out what.
  !
  ! The arguments to the subroutine are the old values of x and y and the new
  ! value of xx. The purpose of the subroutine is to deduce new values of x
  ! and y that correspond to the new value of xx. This is like sliding a bead
  ! on the wire of the hyperbola, hence the name.
  !
  ! The procedure is in three parts. In the first part, the value of 'bend'
  ! is calculated (if possible) from the old x and y. In the second part,
  ! equation (2) above is used to deduce the new value of yy. Thirdly, the
  ! new values of xx and yy are used to calculate the new values of x and y.
  !
  ! There is an extra step which is not part of this required algorithm, but
  ! has been included here for faster processing when the subroutine is used
  ! in context of the program colimplot.

    implicit none

    real(kind=single), intent(in) :: fbend
    real(kind=single), intent(inout) :: x, y
    integer(kind=int8), intent(out) :: status
    real(single) :: bend, newyy, oldxx, newxx, maxyy
real(kind=single) :: newx, newy, oldyy
!, neweryy
    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start:

    status = 0

    if ((x == 0.0 .and. y == 0.0) .or. (x == 1.0 .and. y == 1.0)) then
return
    end if

    if (x < 0.0 .or. x > 1.0 .or. y < 0.0 .or. y > 1.0&
    .or. fbend < -1.0 .or. fbend > 1.0) then
      status = -1
      write(errstr, '(a, e11.4E3, a, e11.4E3, a, f6.3, a)') 'Out of range &
      &arguments ', x, ' ', y, ' ', fbend, ' to subroutine moveBead.'
      call warning('moveBeadOutOfRange')
  return
    end if

    if (y > x) then
      status = -2
      write(errstr, '(a, e11.4E3, a, e11.4E3, a)') 'Arguments ', x, ' '&
      , y, ' to subroutine moveBead are in the wrong order. The first &
      &should be larger than the second.'
      call warning('moveBeadWrongOrder')
  return
    end if

    ! Step 0: calculate the old xx and apply bias to get the new value:
    oldxx = (y + x) / 2.0
!    newxx = (bias(fbend, 2.0 * oldxx - 1.0) + 1.0) / 2.0
call warp(fbend, oldxx, newxx, foreignStatus)
!*** should test foreignStatus!
    if (newxx > 0.5) then
      maxyy = 1.0 - newxx
    else
      maxyy = newxx
    end if

    ! Step I: calculate the curvature of hyperbola from old x, y values:
    bend = (x - y) / (2.0 * x * y - x - y)

    if (bend == 0.0) then ! ie, x = y
      newyy = 0.0
    else
      ! Step II: calculate new value of yy:
      newyy = (1.0 - sqrt(1.0 + 4.0 * bend * bend * newxx * (newxx - 1.0)))&
      / 2.0 / bend

      if (newyy < -maxyy .or. newyy > 0.0) then
        if (newyy < -(maxyy + 0.0001) .or. newyy > 0.0001)&
call error('bugMBny0','bugMBny0')
        newyy = min(0.0, max(-maxyy, newyy))
      end if
    end if

!    ! Step IIa (not really part of the moveBead algorithm, included here in the context of the calling program colimplot): apply bias to newyy:
!    if (maxyy > 0.0) then
!      neweryy = (bias(-clrContrast, 2.0 * (newyy / maxyy) + 1.0) - 1.0)&
!      * maxyy / 2.0
!    else
!      neweryy = 0.0
!    end if
!newyy = neweryy

oldxx = (y + x) / 2.0
oldyy = (y - x) / 2.0
    ! Step III: recalculate x and y:
!    x = newxx - newyy
!    y = newxx + newyy
newx = newxx - newyy
newy = newxx + newyy
x = newx
y = newy

    if (x > 1.0 .or. x < 0.0) then
if (x > 1.0001 .or. x < -0.0001) call error('bugMBx0','bugMBx0')
      x = min(1.0, max(0.0, x))
    end if
    if (y > 1.0 .or. y < 0.0) then
if (y > 1.0001 .or. y < -0.0001) call error('bugMBy0','bugMBy0')
      y = min(1.0, max(0.0, y))
    end if

  end subroutine moveBead

  !---------------------------------------------------------------------
  subroutine finalColourTweak(rgbImage, mask, xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) ::&
      xsize,&
      ysize
    logical, dimension(xsize, ysize), intent(in) :: mask
    real(kind=single), intent(inout) ::&
      rgbImage(xsize, ysize, red:blu)

    integer(kind=int16) :: xi, yi, clr
    real(kind=single) :: bend(red:blu), rgb(red:blu)
integer :: statusPG, pgopen, foreignStatus

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Final colour tweak.')

if (index(testType, 'tweakcheck') == 1) then
  statusPG = pgopen('/xs')
  call pgsvp(0.0, 1.0, 0.0, 1.0)
  call pgwnad(0.0, 3.2, 0.0, 1.1)
  call pgsfs(2)
  call pgsci(1)
  call pgrect(0.0, 1.0, 0.0, 1.0)
  call pgrect(1.1, 2.1, 0.0, 1.0)
  call pgrect(2.2, 3.2, 0.0, 1.0)

  call pgsci(2)
  call pgtext(0.4, 1.05, 'R face')
  call pgsci(3)
  call pgtext(1.5, 1.05, 'G face')
  call pgsci(4)
  call pgtext(2.6, 1.05, 'B face')

  call pgsci(2)
end if

    bend = (colourKnob - 1.0) / (colourKnob + 1.0)
    do xi = 1, xsize
      do yi = 1, ysize
!        rgb(red) = (1.0 + bias(bend(red), 2.0&
!        * rgbImage(xi, yi, red) - 1.0)) / 2.0
!        rgb(grn) = (1.0 + bias(bend(grn), 2.0&
!        * rgbImage(xi, yi, grn) - 1.0)) / 2.0
!        rgb(blu) = (1.0 + bias(bend(blu), 2.0&
!        * rgbImage(xi, yi, blu) - 1.0)) / 2.0
call warp(bend(red), rgbImage(xi, yi, red), rgb(red), foreignStatus)
!*** should test foreignStatus!
call warp(bend(grn), rgbImage(xi, yi, grn), rgb(grn), foreignStatus)
!*** should test foreignStatus!
call warp(bend(blu), rgbImage(xi, yi, blu), rgb(blu), foreignStatus)
!*** should test foreignStatus!

if (index(testType, 'tweakcheck') == 1) then
  call pgmove(real(rgbImage(xi, yi, grn)), real(rgbImage(xi, yi, blu)))
  call pgdraw(real(rgb(grn)), real(rgb(blu)))
  call pgmove(real(1.1 + rgbImage(xi, yi, red)), real(rgbImage(xi, yi, blu)))
  call pgdraw(real(1.1 + rgb(red)), real(rgb(blu)))
  call pgmove(real(2.2 + rgbImage(xi, yi, red)), real(rgbImage(xi, yi, grn)))
  call pgdraw(real(2.2 + rgb(red)), real(rgb(grn)))
end if
        rgbImage(xi, yi, red:blu) = rgb
      end do
    end do

if (index(testType, 'tweakcheck') == 1) then
  call pgclos
end if

    do clr = red, blu
      where (.not. mask) rgbImage(1:xsize, 1:ysize, clr) = nonMaskGreyLevel
    end do

    open(33, file = 'final_image', status ='new')
    write(33, *) rgbImage
    close(33)

  end subroutine finalColourTweak

  !---------------------------------------------------------------------
!  subroutine drawPostscript(rgbImage, mask, xsize, ysize, oldXsize&
subroutine drawPostscript(rgbImage, mask, oldXsize&
  , oldYsize, outFileName)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) ::&
!      xsize,&
!      ysize,&
      oldXsize,&
      oldYsize
!    real(kind=single), intent(in) ::&
!      rgbImage(xsize, ysize, red:blu)
!    logical, intent(in) ::&
!      mask(xsize, ysize)
!logical(bool), intent(in) ::&
logical, intent(in) ::&
  mask(:,:)
real(single), intent(in) ::&
  rgbImage(size(mask,1), size(mask,2), red:blu)

    !character(len=255), intent(in) :: outFileName
    character(*), intent(in) :: outFileName

    integer(int16) ::&
      xsize,&
      ysize,&
      xi,&
      yi,&
      topx,&
      botx,&
      topy,&
      boty
!    integer, parameter :: ioDevNum = 1
integer :: ioDevNum, foreignStatus
    real(kind=single), parameter :: width = 350.0, height = 350.0,&
      xOffset = 100.0, yOffset = 30.0
    real(kind=single) :: deltaX, deltaY, x, y

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Postscript output.')

xsize = size(mask,1)
ysize = size(mask,2)

    if (withMask .and. expandToMask) then
!      call calcCropLimits(mask, xsize, ysize, botx, topx, boty, topy)
call maskArrayLimits(mask, botx, topx, boty, topy, foreignStatus)
    else
      botx = 1
      topx = xsize
      boty = 1
      topy = ysize
    end if

    deltaX = width / real(topx - botx + 1)
    deltaY = height / real(topy - boty + 1)

    ! Write header:

    call getFreeIoUnit(ioDevNum)
    open(unit = ioDevNum, file = outFileName, status = 'replace')
    write(ioDevNum, '(A)') '%!PS-Adobe-3.0'
    write(ioDevNum, '(A)') ''
    write(ioDevNum, '(A)') '/BD{bind def}bind def'
    write(ioDevNum, '(A)') '/LD{load def}BD'
    write(ioDevNum, '(A)') '/ED{exch def}BD'
    write(ioDevNum, '(A)') '/N/newpath LD'
    write(ioDevNum, '(A)') '/M/moveto LD'
    write(ioDevNum, '(A)') '/MS{moveto show}def'
    write(ioDevNum, '(A)') '/D/rlineto LD'
    write(ioDevNum, '(A)') '/F{closepath fill}def'
    write(ioDevNum, '(A)') '/RGB{setrgbcolor}def'
    write(ioDevNum, '(A, F8.4, A)') '/Xwid ', deltaX, ' def'
    write(ioDevNum, '(A, F8.4, A)') '/Ywid ', deltaY, ' def'
    write(ioDevNum, '(A)') '/RGB{setrgbcolor}def'
    write(ioDevNum, '(A)') '/Rect{/Y ED /X ED N X Y M Xwid 0 D 0 Ywid D 0 &
    &Xwid sub 0 D F}def'
    write(ioDevNum, '(A)') '/Hx 100 def'
    write(ioDevNum, '(A)') '/Hx2 {Hx 300 add} def'
    write(ioDevNum, '(A)') '/Hy 700 def'
    write(ioDevNum, '(A)') '/FPx 100 def'
    write(ioDevNum, '(A)') '/FPx2 {FPx 300 add} def'
    write(ioDevNum, '(A)') '/FPy 670 def'
    write(ioDevNum, '(A)') '/PDy 12 def'
    write(ioDevNum, '(A)') ''

    write(ioDevNum, '(A)') '/Helvetica-Bold findfont 12 scalefont setfont'
!    write(ioDevNum, '(A)') '(colimplot v'//version//') Hx 150 add Hy 30 add MS'
    write(ioDevNum, '(A)') '('//trim(nameOf(info))&
    //' v'//trim(versionOf(info))//') Hx 150 add Hy 30 add MS'

    ! Write parameter values:

    write(ioDevNum, '(A)') '/Helvetica-Bold findfont 10 scalefont setfont'
    write(ioDevNum, '(A)') ''
    write(ioDevNum, '(A)') '(Parameter values:) Hx Hy MS'
    if (rebinImagePlease) then
      write(ioDevNum, '(A)') '(rebinimage = T) FPx FPy PDy 0 mul sub MS'
      write(ioDevNum, '(A)') '(  rebinstyle = '//trim(rebinStyle)//') FPx FPy PDy 1 mul sub MS'
      if (rebinStyle == 'tonumber') then
        write(ioDevNum, '(A, I4, A)') '(    newnxbins = ', newNumXbins, ') FPx FPy PDy 2 mul sub MS'
        write(ioDevNum, '(A, I4, A)') '(    newnybins = ', newNumYbins, ') FPx FPy PDy 3 mul sub MS'
      else if (rebinStyle == 'divideby') then
        write(ioDevNum, '(A, F8.3, A)') '(    dividexby = ', divideNumXbinsBy, ') FPx FPy PDy 2 mul sub MS'
        write(ioDevNum, '(A, I8.3, A)') '(    divideyby = ', divideNumYbinsBy, ') FPx FPy PDy 3 mul sub MS'
      else
      end if
    else
      write(ioDevNum, '(A)') '(rebinimage = F) FPx FPy PDy 0 mul sub MS'
    end if
    write(ioDevNum, '(A)')   '(file 1 = '//trim(imageSetName(1))//') FPx FPy PDy 4 mul sub MS'
    if (numSuppliedBands >= 2) then
      write(ioDevNum, '(A)') '(file 2 = '//trim(imageSetName(2))//') FPx FPy PDy 5 mul sub MS'
     if (numSuppliedBands >= 3) then
      write(ioDevNum, '(A)') '(file 3 = '//trim(imageSetName(3))//') FPx FPy PDy 6 mul sub MS'
     end if
    end if
!!!!!
    if (withMask) then
      write(ioDevNum, '(A)') '(withmaskset = T) FPx FPy PDy 7 mul sub MS'
      write(ioDevNum, '(A)') '(maskset = '//trim(maskSetName)//') FPx FPy PDy 8 mul sub MS'
    else
      write(ioDevNum, '(A)') '(withmaskset = F) FPx FPy PDy 7 mul sub MS'
    end if
    write(ioDevNum, '(A)') '(refclrstyle = '//trim(refClrStyle)//') FPx FPy PDy 9 mul sub MS'
    if (refClrStyle == 'internal') then
      write(ioDevNum, '(A)') '(  autowhitestyle = '//trim(autoWhiteStyle)//') FPx FPy PDy 10 mul sub MS'
!      if (autoWhiteStyle == 'mean' .or. autoWhiteStyle == 'rms') then
        write(ioDevNum, '(A, F5.3, A)') '(    cutoff = ', meanCalcRelFluxCutoff,               ') FPx FPy PDy 11 mul sub MS'
!      end if
    else if (refClrStyle == 'bb') then
      write(ioDevNum, '(A)') '(  userwhitestyle = '//trim(userWhiteStyle)//') FPx FPy PDy 10 mul sub MS'
      if (userWhiteStyle == 'kev') then
        write(ioDevNum, '(A, F8.3, A)') '(    whitekev = ', whitekeV,               ') FPx FPy PDy 11 mul sub MS'
      else if (userWhiteStyle == 'angstroms') then
        write(ioDevNum, '(A, F8.3, A)') '(    whiteangstroms = ', whiteAngstroms,   ') FPx FPy PDy 11 mul sub MS'
      else if (userWhiteStyle == 'kelvin') then
        write(ioDevNum, '(A, F8.3, A)') '(    whitemegakelvin = ', whiteMegaKelvin, ') FPx FPy PDy 11 mul sub MS'
      else
      end if
    else
    end if
    write(ioDevNum, '(A, F8.3, A)') '(weirdness = ', weirdness, ') FPx FPy PDy 12 mul sub MS'
    write(ioDevNum, '(A, F8.3, A)') '(heat = ', heat, ') FPx FPy PDy 13 mul sub MS'
    write(ioDevNum, '(A)') '(fluxscalestyle = '//trim(fluxScalingStyle)//') FPx FPy PDy 14 mul sub MS'
    write(ioDevNum, '(A, I3, A)') '(nhotpixels = ', numHotPixels, ') FPx FPy PDy 15 mul sub MS'
    write(ioDevNum, '(A, F5.3, A)') '(gain = ', gain, ') FPx FPy PDy 16 mul sub MS'
    write(ioDevNum, '(A, F5.3, A)') '(bkglevel = ', bkgLevelFrac, ') FPx FPy PDy 17 mul sub MS'
    write(ioDevNum, '(A, F5.3, A)') '(clrcontrast = ', clrContrast, ') FPx FPy PDy 18 mul sub MS'
    write(ioDevNum, '(A, F5.3, A)') '(redamp = ', colourKnob(red), ') FPx FPy PDy 19 mul sub MS'
    write(ioDevNum, '(A, F5.3, A)') '(greenamp = ', colourKnob(grn), ') FPx FPy PDy 20 mul sub MS'
    write(ioDevNum, '(A, F5.3, A)') '(blueamp = ', colourKnob(blu), ') FPx FPy PDy 21 mul sub MS'
    write(ioDevNum, '(A)') '(rgbsysstyle = '//trim(rgbSystemStyle)//') FPx FPy PDy 22 mul sub MS'
    if (preserveAspect) then
      write(ioDevNum, '(A)') '(preserveaspect = T) FPx FPy PDy 23 mul sub MS'
    else
      write(ioDevNum, '(A)') '(preserveaspect = F) FPx FPy PDy 23 mul sub MS'
    end if

    write(ioDevNum, '(A)') '(XMM ID keywords:) Hx2 Hy MS'
    write(ioDevNum, '(A)') '(Instrument = '//trim(instrumeStr)//') FPx2 FPy PDy 0 mul sub MS'
    write(ioDevNum, '(A)') '(Object = '//trim(objectStr)//') FPx2 FPy PDy 1 mul sub MS'
    write(ioDevNum, '(A)') '(Exposure ID = '//trim(expIdStr)//') FPx2 FPy PDy 2 mul sub MS'

    ! Write colour image:

    if (preserveAspect .and. oldXsize .ne. oldYsize) then
!!!!!!!!
    else
      do xi = botx, topx
        x = xOffset + deltaX * (xi - botx)
        do yi = boty, topy
          y = yOffset + deltaY * (yi - boty)

if (rgbImage(xi, yi, red) < 0.0 .or. rgbImage(xi, yi, red) > 1.0&
.or. rgbImage(xi, yi, grn) < 0.0 .or. rgbImage(xi, yi, grn) > 1.0&
.or. rgbImage(xi, yi, blu) < 0.0 .or. rgbImage(xi, yi, blu) > 1.0) then
print *, xi, yi, rgbImage(xi, yi, red), rgbImage(xi, yi, grn), rgbImage(xi, yi, blu)
call error('','')
end if

!if (xi == 202 .and. yi == 105) then
!print *, x, y
!print *, 'in dps', rgbImage(xi, yi, red), rgbImage(xi, yi, grn), rgbImage(xi, yi, blu)
!end if
          write(ioDevNum, '(F6.4, F7.4, F7.4, A)') rgbImage(xi, yi, red)&
          , rgbImage(xi, yi, grn), rgbImage(xi, yi, blu), ' RGB'

          write(ioDevNum, '(F6.2, F7.2, A)') x, y, ' Rect'
        end do
      end do
    end if

    ! Draw box:
    write(ioDevNum, '(A)') '0.0 0.0 0.0 RGB N'
    write(ioDevNum, '(F6.2, F7.2, A)') xOffset, yOffset, ' M'
    write(ioDevNum, '(F6.2, F7.2, A)') width, 0.0, ' D'
    write(ioDevNum, '(F6.2, F7.2, A)') 0.0, height, ' D'
    write(ioDevNum, '(F7.2, F7.2, A)') -width, 0.0&
    , ' D closepath 0.1 setlinewidth stroke'

    write(ioDevNum, '(A)') 'showpage'
    write(ioDevNum, '(A)') '%%EOF'
    close(ioDevNum)
        
  end subroutine drawPostscript

  !---------------------------------------------------------------------
!  subroutine saveToFits(rgbImage, mask, xsize, ysize, outFileName)
subroutine saveToFits(rgbImage, mask, outFileName)
  !---------------------------------------------------------------------

!*** replace by ssclib saveOutput?

    implicit none

!    integer(kind=int16), intent(in) ::&
!      xsize,&
!      ysize
!    real(kind=single), intent(in) ::&
!      rgbImage(xsize, ysize, red:blu)
!    logical, intent(in) ::&
!      mask(xsize, ysize)
!logical(bool), intent(in) ::&
logical, intent(in) ::&
  mask(:,:)
real(single), intent(in) ::&
  rgbImage(size(mask,1), size(mask,2), red:blu)
    !character(len=255), intent(in) :: outFileName
    character(*), intent(in) :: outFileName

    integer(kind=int16) ::&
xsize,&
ysize,&
      topx,&
      botx,&
      topy,&
      boty

    type(DataSetT) :: outSet
    integer, dimension(0:2) :: dimVector
    type(ArrayT) :: outArray
    real(kind=single), dimension(:,:,:), pointer ::&
      arrayDataReal32
    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Writing FITS output.')

xsize = size(mask,1)
ysize = size(mask,2)

    if (withMask .and. expandToMask) then
!      call calcCropLimits(mask, xsize, ysize, botx, topx, boty, topy)
call maskArrayLimits(mask, botx, topx, boty, topy, foreignStatus)
    else
      botx = 1
      topx = xsize
      boty = 1
      topy = ysize
    end if

    outSet = dataset(outFileName, CREATE)
    dimVector(0) = topx - botx + 1
    dimVector(1) = topy - boty + 1
    dimVector(2) = 3
    outArray = addArray(outSet, 'PRIMARY', Real32, dimVector, '', '', 0)
    arrayDataReal32 => real32Array3Data(outArray)
    arrayDataReal32 = rgbImage(botx:topx, boty:topy, :)

    call release(outSet)

  end subroutine saveToFits

  !---------------------------------------------------------------------
!  subroutine saveToPPM(rgbImage, mask, xsize, ysize, outFileName)
subroutine saveToPPM(rgbImage, mask, outFileName)
  !---------------------------------------------------------------------

    implicit none

!    integer(kind=int16), intent(in) ::&
!      xsize,&
!      ysize
!    real(kind=single), intent(inout) ::&
!      rgbImage(xsize, ysize, red:blu)
!    logical, intent(in) ::&
!      mask(xsize, ysize)
!logical(bool), intent(in) ::&
logical, intent(in) ::&
  mask(:,:)
real(single), intent(inout) ::&
  rgbImage(size(mask,1), size(mask,2), red:blu)
    !character(len=255), intent(in) :: outFileName
    character(*), intent(in) :: outFileName

    integer(kind=int16) ::&
      xsize,&
      ysize,&
      xi,&
      yi,&
      topx,&
      botx,&
      topy,&
      boty,&
      redInt, grnInt, bluInt,&
      band, newXsize, newYsize
    integer :: topXpixels, topYpixels, status, pgopen,&
      botVpXpixels, topVpXpixels, botVpYpixels, topVpYpixels
!    integer :: outXsize, outYsize, status, pgopen
    real(kind=single), allocatable ::&
      rebinnedRgbImage(:,:,:)
    integer :: ioDevNum1, ioDevNum2, foreignStatus
    real(single) :: xScale, yScale

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Writing ppm output.')

xsize = size(mask,1)
ysize = size(mask,2)

    if (withMask .and. expandToMask) then
!      call calcCropLimits(mask, xsize, ysize, botx, topx, boty, topy)
call maskArrayLimits(mask, botx, topx, boty, topy, foreignStatus)
    else
      botx = 1
      topx = xsize
      boty = 1
      topy = ysize
    end if

if (withFrame) then
    status = pgopen('frame.gif/gif')
if (status .le. 0) call error('badPgplotDevice', 'PGPLOT did not recognise &
&the device '//trim(pgDev)//' specified by parameter pgdev.')

!call pgqvsz(1, vsLoX, vsHiX, vsLoY, vsHiY) ! get size of view surface
!! in inches
!width = vsHiX - vsLoX
!aspect = (vsHiY - vsLoY) / width
!call pgpap(width / 3.0, aspect)

    call drawVpAndText(int(topx - botx + 1), int(topy - boty + 1)&
    , topXpixels, topYpixels&
    , botVpXpixels, topVpXpixels, botVpYpixels, topVpYpixels)
    call pgclos

    ! Write the pixel locations of the frame left, right, upper and lower
    ! edges to a file that colimchain can read:
    call getFreeIoUnit(ioDevNum2)
    open(unit = ioDevNum2, file = 'chain_coms.dat', status = 'replace')
!     write(ioDevNum2, '(3i6)') int(minVpX * outXsize), int(maxVpX * outXsize)&
!     , outXsize
!     write(ioDevNum2, '(3i6)') int(minVpY * outYsize), int(maxVpY * outYsize)&
!     , outYsize
      write(ioDevNum2, '(3i6)') botVpXpixels, topVpXpixels, topXpixels
      write(ioDevNum2, '(3i6)') botVpYpixels, topVpYpixels, topYpixels
    close(ioDevNum2)

    ! The following is done because we want to preserve (as near as possible)
    ! the aspect ratio of the image:
!    xScale = outXsize * (maxVpX - minVpX) / real(topx - botx + 1)
!    yScale = outYsize * (maxVpY - minVpY) / real(topy - boty + 1)
    xScale = (topVpXpixels - botVpXpixels) / real(topx - botx + 1)
    yScale = (topVpYpixels - botVpYpixels) / real(topy - boty + 1)
    newXsize = int((topx - botx + 1) * min(xScale, yScale))
    newYsize = int((topy - boty + 1) * min(xScale, yScale))

    ! Rebin the colour image so that it fits neatly into the frame:
    allocate(rebinnedRgbImage(newXsize, newYsize, red:blu))
    do band = red, blu
      call imageRebinner(rgbImage(botx:topx, boty:topy, band)&
      , rebinnedRgbImage(:, :, band))
    end do
    where(rebinnedRgbImage > 1.0) rebinnedRgbImage = 1.0

    call getFreeIoUnit(ioDevNum1)
    open(unit = ioDevNum1, file = outFileName, status = 'replace')
      write(ioDevNum1, '(A)') 'P3'
      write(ioDevNum1, '(i5)') newXsize
      write(ioDevNum1, '(i5)') newYsize
      write(ioDevNum1, '(i3)') 255

!      do yi = 1, newYsize
      do yi = newYsize, 1, -1
        do xi = 1, newXsize
          redInt = min(255, floor(rebinnedRgbImage(xi, yi, red) * 256.0))
          grnInt = min(255, floor(rebinnedRgbImage(xi, yi, grn) * 256.0))
          bluInt = min(255, floor(rebinnedRgbImage(xi, yi, blu) * 256.0))
          write(ioDevNum1, '(3i4)') redInt, grnInt, bluInt
        end do
      end do
    close(ioDevNum1)

else
    call getFreeIoUnit(ioDevNum1)
    open(unit = ioDevNum1, file = outFileName, status = 'replace')
      write(ioDevNum1, '(A)') 'P3'
      write(ioDevNum1, '(i5)') topx - botx + 1
      write(ioDevNum1, '(i5)') topy - boty + 1
      write(ioDevNum1, '(i3)') 255

!      do yi = boty, topy
      do yi = topy, boty, -1
        do xi = botx, topx
          redInt = min(255, floor(rgbImage(xi, yi, red) * 256.0))
          grnInt = min(255, floor(rgbImage(xi, yi, grn) * 256.0))
          bluInt = min(255, floor(rgbImage(xi, yi, blu) * 256.0))
          write(ioDevNum1, '(3i4)') redInt, grnInt, bluInt
        end do
      end do
    close(ioDevNum1)
end if

  end subroutine saveToPPM

  !---------------------------------------------------------------------
  subroutine toPolarCoord(inImage, outImage, xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) ::&
      xsize,&
      ysize
    real(kind=single), intent(in) ::&
      inImage(xsize, ysize, red:blu)
    real(kind=single), intent(out) ::&
      outImage(xsize, ysize, red:blu)

    integer(kind=int16) ::&
      xi,&
      yi,&
      clr,&
      startClr,&
      clrOfMin,&
      clrOfMed,&
      clrOfMax
    real(kind=single) ::&
      redAngle,&
      cosRedAngle,&
      sinRedAngle,&
      minVal,&
      medVal,&
      maxVal,&
      red0,&
      red1,&
      red2,&
      grn0,&
      grn1,&
      grn2,&
      blu0,&
      blu1,&
      blu2

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Conversion of RGB points to polar coord.')

!    redAngle = atan(root2)
redAngle = atan(RootTwo)
    cosRedAngle = cos(redAngle)
    sinRedAngle = sin(redAngle)

    do xi = 1, xsize
      PIXEL: do yi = 1, ysize
        maxVal = inImage(xi, yi, red)
        clrOfMax = red
        do clr = grn, blu
          if (maxVal < inImage(xi, yi, clr)) then
            maxVal = inImage(xi, yi, clr)
            clrOfMax = clr
          end if
        end do

        if (maxVal == 0.0) then
          outImage(xi, yi, red:blu) = 0.0
      cycle PIXEL
        end if

        do clr = red, blu
          if (clr .ne. clrOfMax) then
            startClr = clr
        exit
          end if
        end do

        medVal = inImage(xi, yi, startClr)
        clrOfMed = startClr
        do clr = startClr + 1, blu
          if (clr .ne. clrOfMax .and. medVal < inImage(xi, yi, clr)) then
            medVal = inImage(xi, yi, clr)
            clrOfMed = clr
          end if
        end do

        do clr = red, blu
          if (clr .ne. clrOfMax .and. clr .ne. clrOfMed) then
            clrOfMin = clr
        exit
          end if
        end do
        minVal = inImage(xi, yi, clrOfMin)

        outImage(xi, yi, red) = maxVal
        outImage(xi, yi, blu) = minVal / maxVal

        red0 = inImage(xi, yi, red)
        grn0 = inImage(xi, yi, grn)
        blu0 = inImage(xi, yi, blu)

        ! Rotate about blu axis by 45 deg, red -> grn:
!***        red1 = red0 / root2 - grn0 / root2
!***        grn1 = grn0 / root2 + red0 / root2
red1 = red0 / RootTwo - grn0 / RootTwo
grn1 = grn0 / RootTwo + red0 / RootTwo
        blu1 = blu0

        ! Rotate about red axis by atan(sqrt(2)), grn -> blu:
        red2 = red1
        grn2 = grn1 * cosRedAngle - blu1 * sinRedAngle
        blu2 = blu1 * cosRedAngle + grn1 * sinRedAngle

        ! Looking down the blu axis onto the colour cube one now sees it
        ! tilted so that the 3-symmetry of the cube is exposed: the white
        ! point appears superimposed (ie (red2, grn2) the same for both
        ! points) on the black.

!        outImage(xi, yi, grn) = atan2(red2, grn2) / twoPi
outImage(xi, yi, grn) = atan2(red2, grn2) / TwoPi
! from constants module
        if (outImage(xi, yi, grn) < 0.0) then
          outImage(xi, yi, grn) = 1.0 + outImage(xi, yi, grn)
        end if

        do clr = red, blu
          outImage(xi, yi, clr) = min(max(outImage(xi, yi, clr), 0.0), 1.0)
        end do
      end do PIXEL
    end do

  end subroutine toPolarCoord

  !---------------------------------------------------------------------
  subroutine assignColourValues(rgbImage, xsize, ysize, maxNumBranches&
  , indexImage, lastClrIndex, clrValuesRet)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) ::&
      maxNumBranches,&
      xsize,&
      ysize
    real(kind=single), dimension(xsize, ysize, red:blu), intent(in) ::&
      rgbImage
    real, dimension(xsize, ysize), intent(out) ::&
      indexImage
    integer, intent(out) ::&
      lastClrIndex
    !real(kind=single),dimension(minClrIndex:maxClrIndex,red:blu),intent(out)::&
    real(kind=single), intent(out)::&
      clrValuesRet(:,:)

    real(kind=single), allocatable :: clrValues(:,:)	! used here locally, which will be passed to clrValuesRet as it is in the end or return of this subroutine.

    integer(kind=int16) ::&
      numBranches,&
      branchMap(xsize, ysize),&
      clr,&
      branch,&
      biggestLeaf,&
      axisOfBiggest,&
      clrIndex
    type(treeType), dimension(maxNumBranches) ::&
      kdTree
    real(kind=single) ::&
      maxDimValue,&
      splitValue,&
      tempArray(xsize, ysize),&
      maxValue,&
      minValue
!character(len=1) :: char

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Assigning the colour values.')

    allocate(clrValues(minClrIndex:maxClrIndex,red:blu))

    kdTree%empty = .true. ! vector op, default
    numBranches = 1

    ! Set up first branch/cell:
    do clr = red, blu
      kdTree(1)%max(clr) = maxval(rgbImage(1:xsize, 1:ysize, clr))
      kdTree(1)%min(clr) = minval(rgbImage(1:xsize, 1:ysize, clr))
    end do
!***** can do the following instead??
!    kdTree(1)%max = maxval(rgbImage)
!    kdTree(1)%min = minval(rgbImage)
    kdTree(1)%empty = .false.
    kdTree(1)%isLeaf = .true.
    kdTree(1)%clrIndex = minClrIndex
    lastClrIndex = minClrIndex

    branchMap = minClrIndex ! array op

    ! Construct rest of tree:
    do while (numBranches < maxNumBranches)
!print *, 'blrwrwrwrr!!', numBranches
!if (floor((numBranches-1)/20.0) == (numBranches-1)/20.0) then
!  print *, (numBranches-1)/2, '/', (maxNumBranches-1)/2, 'colours done.'
!end if

      ! Find leaf/cell with largest dimension:
      maxDimValue = 0.0 ! size of largest dimension of the cell.
      do branch = 1, numBranches
        if (kdTree(branch)%isLeaf) then
          do clr = red, blu
            if (kdTree(branch)%max(clr) - kdTree(branch)%min(clr)&
            > maxDimValue) then
              maxDimValue = kdTree(branch)%max(clr) - kdTree(branch)%min(clr)
              biggestLeaf = branch
              axisOfBiggest = clr
            end if
          end do
        end if
      end do

    if (maxDimValue .le. 0.0) exit ! because there is therefore now only
    ! one point in each branch.

      ! Split the biggest leaf/cell:
      kdTree(biggestLeaf)%isLeaf = .false.

      splitValue = (kdTree(biggestLeaf)%max(axisOfBiggest)&
      + kdTree(biggestLeaf)%min(axisOfBiggest)) / 2.0

      kdTree(numBranches + 1)%isLeaf = .true.
      kdTree(numBranches + 1)%empty  = .false.
      kdTree(numBranches + 1)%min = kdTree(biggestLeaf)%min
      kdTree(numBranches + 1)%max = kdTree(biggestLeaf)%max
      kdTree(numBranches + 1)%max(axisOfBiggest) = splitValue
      kdTree(numBranches + 1)%clrIndex = kdTree(biggestLeaf)%clrIndex
      kdTree(numBranches + 2)%isLeaf = .true.
      kdTree(numBranches + 2)%empty  = .false.
      kdTree(numBranches + 2)%min = kdTree(biggestLeaf)%min
      kdTree(numBranches + 2)%max = kdTree(biggestLeaf)%max
      kdTree(numBranches + 2)%min(axisOfBiggest) = splitValue
      kdTree(numBranches + 2)%clrIndex = lastClrIndex + 1
      lastClrIndex = lastClrIndex + 1

      where (branchMap == kdTree(biggestLeaf)%clrIndex&
      .and. rgbImage(1:xsize, 1:ysize, axisOfBiggest) > splitValue)
        branchMap = kdTree(numBranches + 2)%clrIndex
      end where

      ! Shrink the two daughter cells to fit snugly around the points
      ! they contain:
      do branch = numBranches + 1, numBranches + 2
        do clr = red, blu
          where (branchMap == kdTree(branch)%clrIndex)
            tempArray = rgbImage(1:xsize, 1:ysize, clr)
          elsewhere
            tempArray = 0.0
          end where
          maxValue = maxval(tempArray)
          where (branchMap .ne. kdTree(branch)%clrIndex)
            tempArray = maxValue
          end where
          minValue = minval(tempArray)
          kdTree(branch)%max(clr) = maxValue
          kdTree(branch)%min(clr) = minValue
        end do
      end do

      numBranches = numBranches + 2
    end do

    do branch = 1, numBranches
      if (kdTree(branch)%isLeaf) then
        clrIndex = kdTree(branch)%clrIndex

        if (kdTree(branch)%max(red) == 1.0&
        .and. kdTree(branch)%max(grn) == 1.0&
        .and. kdTree(branch)%max(blu) == 1.0) then
        ! make white white:
          clrValues(clrIndex, red:blu) = 1.0

!if (branch == 168) then
!print *, kdTree(branch)%min(red), kdTree(branch)%min(grn), !kdTree(branch)%min(blu)
!end if

        else if (kdTree(branch)%min(red) == 0.0&
        .and. kdTree(branch)%min(grn) == 0.0&
        .and. kdTree(branch)%min(blu) == 0.0) then
        ! make black black:
          clrValues(clrIndex, red:blu) = 0.0

        else
          clrValues(clrIndex, red:blu) = (kdTree(branch)%max(red:blu)&
          + kdTree(branch)%min(red:blu)) / 2.0
        end if
      end if
    end do

    indexImage = real(branchMap) + 0.5 ! array op
!!!!!! Could probably do away with branchMap and just use indexImage.

    clrValuesRet =  clrValues	! The start indices are different.

  end subroutine assignColourValues

  !---------------------------------------------------------------------
  subroutine fromPolarCoord(imageArray, outArray)
  !---------------------------------------------------------------------

    implicit none

    real(kind=single), dimension(minClrIndex:maxClrIndex, red:blu) ::&
      imageArray, outArray	!!! Check!!!! ??????
    integer(kind=int16) ::&
      clr,&
      clrOfMin,&
      clrOfMed,&
      clrOfMax,&
      i
    real(kind=single) ::&
      angleNorm,&
      angle,&
      minVal,&
      medVal,&
      maxVal

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Converting colour values from polar &
    &to cartesian coords.')

    do i = minClrIndex, maxClrIndex

      maxVal = imageArray(i, red)
      minVal = imageArray(i, blu) * maxVal

      angleNorm = imageArray(i, grn)
      ! units such that full circle = 1.

      if (angleNorm < 1/3.0) then
        clrOfMax = red
        if (angleNorm < 1/6.0) then
          clrOfMed = grn
          clrOfMin = blu
        else
          angleNorm = 1/3.0 - angleNorm
          clrOfMed = blu
          clrOfMin = grn
        end if
      else if (angleNorm < 2/3.0) then
        angleNorm = angleNorm - 1/3.0
        clrOfMax = blu
        if (angleNorm < 1/6.0) then
          clrOfMed = red
          clrOfMin = grn
        else
          angleNorm = 1/3.0 - angleNorm
          clrOfMed = grn
          clrOfMin = red
        end if
      else
        angleNorm = angleNorm - 2/3.0
        clrOfMax = grn
        if (angleNorm < 1/6.0) then
          clrOfMed = blu
          clrOfMin = red
        else
          angleNorm = 1/3.0 - angleNorm
          clrOfMed = red
          clrOfMin = blu
        end if
      end if

!***      angle = (angleNorm - 1/12.0) * twoPi
angle = (angleNorm - 1/12.0) * TwoPi
! from constants module

!***      medVal = (maxVal + minVal) / 2.0 - (maxVal - minVal) * root3&
!***      * tan(angle) / 2.0
medVal = (maxVal + minVal) / 2.0 - (maxVal - minVal) * RootThree&
* tan(angle) / 2.0

      outArray(i, clrOfMin) = minVal
      outArray(i, clrOfMed) = medVal
      outArray(i, clrOfMax) = maxVal

      do clr = red, blu
        if (outArray(i, clr) < 0.0)&
          outArray(i, clr) = 0.0
        if (outArray(i, clr) > 1.0)&
          outArray(i, clr) = 1.0
      end do
    end do

  end subroutine fromPolarCoord

  !---------------------------------------------------------------------
  subroutine outputToPgplot(aveClrValues, minClrIndex, maxClrIndex&
!  , lastClrIndex, indexImage, mask, xsize, ysize)
, lastClrIndex, mask, indexImage)
  !---------------------------------------------------------------------

    implicit none

    integer,       intent(in) :: minClrIndex, maxClrIndex, lastClrIndex
    real(single),  intent(in) :: aveClrValues(minClrIndex:maxClrIndex, red:blu)
!    logical(bool), intent(in) :: mask(:,:)
    logical,       intent(in) :: mask(:,:)
    real,          intent(in) :: indexImage(size(mask,1),size(mask,2))

    integer(int16) ::&
      minx, maxx, miny, maxy
!, xi, yi
    integer :: status, foreignStatus, pgopen
    real :: vsLoX, vsHiX, vsLoY, vsHiY, width, aspect
    character(len=255) :: pgFileDev
    integer(int16) ::&
      xsize, ysize

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Writing output.')

    xsize = size(mask,1)
    ysize = size(mask,2)

if (withMask .and. expandToMask) then
  call maskArrayLimits(mask, minx, maxx, miny, maxy, foreignStatus)
!*** should test foreignStatus!

!!vvvvv this is clumsy. Use Clive's way?
!  minx = 1
!  OUTER1: do xi = 2, xsize
!    do yi = 1, ysize
!  if (mask(xi, yi)) exit OUTER1
!    end do
!    minx = xi
!  end do OUTER1
!  maxx = xsize
!  OUTER2: do xi = xsize - 1, 1, -1
!    do yi = 1, ysize
!  if (mask(xi, yi)) exit OUTER2
!    end do
!    maxx = xi
!  end do OUTER2
!  miny = 1
!  OUTER3: do yi = 2, ysize
!    do xi = 1, xsize
!  if (mask(xi, yi)) exit OUTER3
!    end do
!    miny = yi
!  end do OUTER3
!  maxy = ysize
!  OUTER4: do yi = ysize - 1, 1, -1
!    do xi = 1, xsize
!  if (mask(xi, yi)) exit OUTER4
!    end do
!    maxy = yi
!  end do OUTER4
!!^^^^^

else
  minx = 1
  maxx = xsize
  miny = 1
  maxy = ysize
end if

!    call plotNow_20031023(indexImage, xsize, ysize, minx, maxx, miny, maxy&
!    , aveClrValues, minClrIndex, maxClrIndex, lastClrIndex)
call plotNow(indexImage(minx:maxx, miny:maxy)&
, aveClrValues, minClrIndex, maxClrIndex, lastClrIndex)

    call pgclos

    if (doThumbnail .and. (pgDev == '/ps' .or. pgDev == '/gif')) then
!print *, 'doThumbnail', doThumbnail
      call message(AppMsg, Verbose, 'Writing thumbnail image.')

      pgFileDev = trim(outFileName)//'.small'//trim(pgDev)
      status = pgopen(pgFileDev)
if (status .le. 0) call error('badPgplotDevice', 'PGPLOT did not recognise &
&the device '//trim(pgFileDev)//'.')
      call pgqvsz(1, vsLoX, vsHiX, vsLoY, vsHiY) ! get size of view surface
      ! in inches
      width = vsHiX - vsLoX
      aspect = (vsHiY - vsLoY) / width
      call pgpap(width / 10.0, aspect)

!      call plotNow_20031023(indexImage, xsize, ysize, minx, maxx, miny, maxy&
!      , aveClrValues, minClrIndex, maxClrIndex, lastClrIndex)
call plotNow(indexImage(minx:maxx, miny:maxy)&
, aveClrValues, minClrIndex, maxClrIndex, lastClrIndex)

      call pgclos
    end if

  end subroutine outputToPgplot

  !---------------------------------------------------------------------
  subroutine plotNow(indexImage&
  , aveClrValues, minClrIndex, maxClrIndex, lastClrIndex)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) ::&
      indexImage(:,:)
    integer, intent(in) ::&
      minClrIndex,&
      maxClrIndex,&
      lastClrIndex
    real(single), intent(in) ::&
      aveClrValues(minClrIndex:maxClrIndex, red:blu)

    integer ::&
      clrIndex,&
      minx,&
      maxx,&
      miny,&
      maxy

    !-------------------------------------------------------------------
    ! Start:

    minx = 1.0
    maxx = size(indexImage, 1)
    miny = 1.0
    maxy = size(indexImage, 2)

    call pgwnad(minx-0.5, maxx+0.5, miny-0.5, maxy+0.5) ! note the integer
    ! min*, max* get converted to the necessary reals by adding +-0.5.

    call drawVpAndText(maxx, maxy)

    call pgscir(minClrIndex, lastClrIndex)
    do clrIndex = minClrIndex, lastClrIndex
      call pgscr(clrIndex&
      , real(aveClrValues(clrIndex, red))&
      , real(aveClrValues(clrIndex, grn))&
      , real(aveClrValues(clrIndex, blu)))
    end do
    call pgimag(indexImage, maxx, maxy, 1, maxx, 1&
    , maxy, real(minClrIndex), real(lastClrIndex + 1), trMatrix)

  end subroutine plotNow

  !---------------------------------------------------------------------
  subroutine drawVpAndText(maxx, maxy, topXpixels, topYpixels&
  , botVpXpixels, topVpXpixels, botVpYpixels, topVpYpixels)
  !---------------------------------------------------------------------
  ! Code to place text etc as near as possible to the way it is on the
  ! output of task inplot:

    implicit none

    integer, intent(in) ::&
      maxx,&
      maxy
    integer, intent(out), optional ::&
      topXpixels, topYpixels,&
      botVpXpixels, topVpXpixels, botVpYpixels, topVpYpixels

    character(80) ::&
      legend
!    real ::&
!      x,&
!      y,&
!      dy,&
!      csize
    integer ::&
!      font,&
      minx, miny
    real :: x1, x2, y1, y2

    !-------------------------------------------------------------------
    ! Start:

minx = 0.0
miny = 0.0

    call pgqvsz(3, x1, x2, y1, y2)
    if (present(topXpixels)) topXpixels = int(x2 - x1 + 1)
    if (present(topYpixels)) topYpixels = int(y2 - y1 + 1)

    call pgsvp(minVpX, maxVpX, minVpY, maxVpY)
    call pgwnad(0.5, maxx+0.5, 0.5, maxy+0.5)
    call pgqvp(3, x1, x2, y1, y2)
    if (present(botVpXpixels)) botVpXpixels = int(x1)
    if (present(topVpXpixels)) topVpXpixels = int(x2)
    if (present(botVpYpixels)) botVpYpixels = int(y1)
    if (present(topVpYpixels)) topVpYpixels = int(y2)

    call pgsch(0.7)
    call pgmtxt("T", 1.4, 0.5, 0.5, 'False-colour xray image.')
    write(legend, "(a)") trim(nameOf(info))//" v"//trim(versionOf(info))
    call pgmtxt("T", 1.2, 1.0, 0.0, trim(legend))
    call pgscf(2)                 ! better quality Roman font
    call pgbox('BCINT', 0.0, 0, 'BCINT', 0.0, 0)
    call pglab('X', 'Y', " ")

call plot_headers(minx, maxx, miny, maxy)

    call pgiden

  end subroutine drawVpAndText

!-----------------------------------------------------------------------
! plot_headers: shows FITS headers at top of plot
!-----------------------------------------------------------------------
subroutine plot_headers(minx, maxx, miny, maxy)
integer, intent(in)       :: minx, maxx, miny, maxy ! world coordinate limits
!
integer :: font
!character*30 :: string
character*30 :: pixels
real :: x, y, dy, csize
!
x = 1.05
y = 0.95
dy = 0.05
call pgqcf(font)
call pgscf(1)
call pgqch(csize)
call pgsch(0.5)
call pgmtxt("RV", x, y, 0.0, trim(instrumeStr))

y = y - dy
if(filterStr /= " ") then
   call pgmtxt("RV", x, y, 0.0, trim(filterStr)//' filter')
end if
y = y - dy
call pgmtxt("RV", x, y, 0.0, trim(objectStr))
y = y - dy
call pgmtxt("RV", x, y, 0.0, trim(observerStr))
y = y - dy
call pgmtxt("RV", x, y, 0.0, dateobsStr(1:16))
y = y - dy
if(exposureStr /= " ") then
   call pgmtxt("RV", x, y, 0.0, trim(exposureStr)//' sec')
end if
!
!pixels = "(" // trim(itoc(minx)) // ":" // trim(itoc(maxx)) // "," // &
!                trim(itoc(miny)) // ":" // trim(itoc(maxy)) // ") pixels"
pixels = "(" // trim(intToStr(minx, leftJustify=.true._bool))&
  // ":" // trim(intToStr(maxx, leftJustify=.true._bool))&
  // "," // trim(intToStr(miny, leftJustify=.true._bool))&
  // ":" // trim(intToStr(maxy, leftJustify=.true._bool)) // ") pixels"

y = y - 1.5 * dy
call pgmtxt("RV", x, y, 0.0, trim(pixels))
!
call pgscf(font)
call pgsch(csize)
return
end subroutine plot_headers


end module colimplot_mod
